import os
from typing import *

import torch
import transformers

from realhf.api.core.model_api import ReaLModelConfig, register_hf_family

from .llama import (
    convert_state_dict_llama,
    llama_embedding_layer_names,
    llama_output_head_param_name,
    llama_transformer_block_param_name,
    to_llama_state_dict,
)


def convert_config_qwen2(
    hf_config: transformers.Qwen2Config,
) -> ReaLModelConfig:
    return ReaLModelConfig(
        n_layers=hf_config.num_hidden_layers,
        n_kv_heads=hf_config.num_key_value_heads,
        hidden_dim=hf_config.hidden_size,
        n_q_heads=hf_config.num_attention_heads,
        head_dim=hf_config.hidden_size // hf_config.num_attention_heads,
        intermediate_dim=hf_config.intermediate_size,
        vocab_size=hf_config.vocab_size,
        n_positions=hf_config.max_position_embeddings,
        embd_pdrop=0.0,
        attn_pdrop=(
            hf_config.attention_dropout
            if hasattr(hf_config, "attention_dropout")
            else 0.1
        ),
        layer_norm_epsilon=hf_config.rms_norm_eps,
        activation_function=hf_config.hidden_act,
        use_attention_bias=True,
        use_attn_proj_bias=False,
        scale_attn_by_inverse_layer_idx=False,
        layer_norm_type="rms",
        mlp_type="llama",
        apply_rotary=True,
        rotary_base=hf_config.rope_theta,
        rotary_interleaved=False,
    )


def convert_config_back_qwen2(
    config: ReaLModelConfig,
) -> transformers.Qwen2Config:
    return transformers.Qwen2Config(
        vocab_size=config.vocab_size,
        hidden_size=config.hidden_dim,
        intermediate_size=config.intermediate_dim,
        num_hidden_layers=config.n_layers,
        num_key_value_heads=config.n_kv_heads,
        num_attention_heads=config.n_q_heads,
        max_position_embeddings=config.n_positions,
        rms_norm_eps=config.layer_norm_epsilon,
        hidden_act=config.activation_function,
        attention_dropout=config.attn_pdrop,
        rope_theta=config.rotary_base,
    )


def qwen2_config_maker():
    hf_config = transformers.Qwen2Config(
        vocab_size=200,
        hidden_size=128,
        intermediate_size=200,
        num_hidden_layers=8,
        num_attention_heads=8,
        num_key_value_heads=8,
        hidden_act="silu",
        rms_norm_eps=1e-5,
    )
    return convert_config_qwen2(hf_config)


register_hf_family(
    name="qwen2",
    hf_cls_name="Qwen2ForCausalLM",
    config_from_hf_converter=convert_config_qwen2,
    config_to_hf_converter=convert_config_back_qwen2,
    sd_from_hf_converter=convert_state_dict_llama,
    sd_to_hf_converter=to_llama_state_dict,
    embedding_param_names=llama_embedding_layer_names,
    tblock_param_names=llama_transformer_block_param_name,
    head_param_names=llama_output_head_param_name,
    real_config_maker=qwen2_config_maker,
)



import contextlib
import dataclasses
import functools
from collections import defaultdict
from typing import *

import numpy as np
from omegaconf import MISSING

import realhf.base.logging as logging
from realhf.api.core.config import Dataset
from realhf.api.core.dfg import MFCDef, ModelInterfaceType, OffloadHook, SyncParamHook
from realhf.api.core.system_api import *
from realhf.api.quickstart.device_mesh import (
    AllocationConfig,
    DeviceMesh,
    RPCAllocation,
    make_device_mesh_from_name,
)
from realhf.api.quickstart.model import ModelTrainEvalConfig, ParallelismConfig
from realhf.experiments.common.check import *
from realhf.experiments.common.utils import *
from realhf.search_engine.search import search_rpc_allocations

logger = logging.getLogger("CommonExperimentConfig", "colored")


@dataclasses.dataclass
class CommonExperimentConfig(Experiment):
    """The common config for quickstart experiments.

    All members can be changed by the user in the command line,
    e.g.,

    .. code-block:: shell

        $ python3 -m realhf.apps.quickstart sft trial_name=my_trial seed=42 exp_ctrl.save_freq_steps=10 ...

    The above command changes the ``trial_name``, the ``seed`` attribute,
    and the ``save_freq_steps`` attribute of the ``exp_ctrl`` attribute
    of this class.

    ``recover_mode`` is one of the followings\:

    - ``auto``\: automatically recover the last failed run.

    - ``save``\: save recover states if any error occurs.

    - ``resume``\: resume from saved recover states and save states if fail again.

    - ``disabled``\: do nothing but raise error when error occurs.

    If you don't know how ReaL's recovery works, set it to be ``disabled``.
    Normal checkpointing is usually sufficient for most cases.

    ``allocation_mode`` is one of the followings\:

    - ``manual``\: manually allocate resources with given commandline options.

    - ``search``\: allocate resources and configure parallel strategies with the search engine.

    - ``heuristic``\: allocate resources and configure parallel strategies with heuristic strategies previously obtained by search.

    - ``pipe_data``\: identical parallelization (like DSChat) with pipe+data parallelism. A world size under 8 will use data parallelism only.

    - ``pipe_model``\: identical parallelization (like DSChat) with pipe+model parallelism. A world size under 8 will use tensor-model parallelism only.

    :param experiment_name: Name of the experiment. Arbitrary string without "-" and "/", e.g., ``ultra-chat-llama``.
    :type experiment_name: str
    :param trial_name: Name of the trial. Arbitrary string without "-" and "/", e.g., ``lr1e-3wd0.05``.
    :type trial_name: str
    :param mode: Experiment launching mode. "local", "ray", or "slurm" are supported.
        "ray" mode requires launching the ray cluster with CLI.
        "slurm" mode requires the pyxis plugin with enroot container enabled.
        The "local" mode implies ``n_nodes=1``.
    :type mode: str
    :param debug: Whether to run in the debug mode.
        Setting to `False` will disable all assertions, which will be unsafe but faster.
    :type debug: bool
    :param partition: The slurm partition to run the experiment.
    :type partition: str
    :param wandb_mode: The mode of wandb. Currently the wandb logging is not supported.
    :type wandb_mode: str
    :param image_name: The name of the docker image used by the controller.
        Only used in the slurm mode.
    :type image_name: str or None
    :param recover_mode: The recover mode. See above.
    :type recover_mode: str
    :param recover_retries: The number of retries for recovery.
        Only effective when recover_mode is "auto".
    :type recover_retries: int
    :param ignore_worker_error: Whether to ignore errors raised by
        workers during runtime. Please do not set it to be True unless
        the user is sure that some error is ignorable.
        Only effective when recover_mode is "disabled".
    :type ignore_worker_error: bool
    :param allocation_mode: Mode of GPU parallel strategy allocation. See above.
    :type allocation_mode: str
    :param allocation_use_cache: Whether to use cache in allocation search.
        Only effective when allocation_mode=="search"
        and cache is available in the log dir of current experiment
        name and trial.
    :type allocation_use_cache: bool
    :param n_nodes: Number of nodes to run the experiment.
    :type n_nodes: int
    :param n_gpus_per_node: Number of GPUs per node.
        Thus, the total number of GPUs will be ``n_nodes * n_gpus_per_node``.
        ReaL supports a world size of 1, 2, 4, 8, ... within a single node,
        or multiple nodes with the same number of GPUs.
    :type n_gpus_per_node: int
    :param nodelist: Nodelist for the distributed setting in SLURM nodelist format.
        Required for the ``manual`` allocation mode.
        For several GPUs on a single node, it should be like "NODE01:0,1,2,3",
        meaning that using the first 4 GPUs on the ``NODE01`` node.
        For several complete nodes, it should be like "NODE[01-02,03,07],COM08",
        meaning that using all GPUs for these nodes: [NODE01, NODE02, NODE03, NODE07, COM08].
    :type nodelist: str or None
    :param seed: Random seed.
    :type seed: int
    :param exp_ctrl: The save and evaluation control of the experiment.
    :type exp_ctrl: ExperimentSaveEvalControl
    """

    experiment_name: str = MISSING
    trial_name: str = MISSING
    mode: str = dataclasses.field(
        metadata={"choices": ["slurm", "local", "ray"]}, default="slurm"
    )
    debug: bool = True
    partition: str = "dev"
    wandb_mode: str = "disabled"
    image_name: Optional[str] = None
    recover_mode: str = "disabled"
    recover_retries: int = 1
    ignore_worker_error: bool = False
    allocation_mode: str = "pipe_model"
    allocation_use_cache: bool = False
    n_nodes: int = 1
    n_gpus_per_node: int = 8
    nodelist: Optional[str] = None
    seed: int = 1
    exp_ctrl: ExperimentSaveEvalControl = dataclasses.field(
        default_factory=ExperimentSaveEvalControl
    )

    @property
    def models(self) -> Dict[str, ModelTrainEvalConfig]:
        raise NotImplementedError(f"models is not implemented in {self.__class__}")

    @property
    def rpcs(self) -> Dict[str, MFCDef]:
        raise NotImplementedError(f"rpcs is not implemented in {self.__class__}")

    @property
    def datasets(self) -> List[Dataset]:
        return []

    @property
    def eval_datasets(self) -> List[Dataset]:
        return None

    @property
    def eval_dataloader(self) -> DataLoader:
        return DataLoader("packed_eval", args=dict(batch_size=128))

    @property
    def tokenizer_name_or_path(self) -> str:
        raise NotImplementedError(
            f"tokenizer_name_or_path is not implemented in {self.__class__}"
        )

    @property
    def max_prompt_len(self) -> int:
        return None

    @property
    def search_kwargs(self) -> Dict[str, Any]:
        return {}

    @property
    def allocations(self) -> Dict[str, AllocationConfig]:
        return {}

    @property
    def global_device_mesh(self) -> DeviceMesh:
        return DeviceMesh(
            n_nodes=self.n_nodes,
            n_gpus_per_node=self.n_gpus_per_node,
            mapping=np.ones((self.n_nodes, self.n_gpus_per_node), dtype=np.int32),
            global_mesh_name=self.nodelist,
            name=self.nodelist,
        )

    def _heuristic_rpc_allocation(self) -> List[RPCAllocation]:
        raise NotImplementedError(
            f"_heuristic_rpc_allocation is not implemented in {self.__class__}"
        )

    def _search(self):
        # called in both api.main and controller
        gradient_checkpointing = any(
            model.gradient_checkpointing for model in self.models.values()
        )
        rpc_allocs: List[RPCAllocation] = search_rpc_allocations(
            device_mesh=self.global_device_mesh,
            rpcs=list(self.rpcs.values()),
            gradient_checkpointing=gradient_checkpointing,
            use_cache=self.allocation_use_cache,
            **self.search_kwargs,
        )
        return rpc_allocs

    def scheduling_setup(self) -> ExperimentScheduling:
        return ExperimentScheduling(
            master_worker=TasksGroup(
                count=1,
                scheduling=Scheduling.master_worker_default(
                    cpu=4, mem=20000, nodelist=self.nodelist
                ),
            ),
            model_worker=TasksGroup(
                count=self.n_nodes * self.n_gpus_per_node,
                scheduling=Scheduling.model_worker_default(
                    cpu=4,
                    gpu=1,
                    gpu_type="tesla",
                    mem=100000,
                    nodelist=self.nodelist,
                ),
            ),
        )

    def initial_setup(self) -> ExperimentConfig:
        if self.allocation_mode == "manual" and self.nodelist is None:
            logger.warning(
                "Warning: Nodelist is not set in manual allocation mode, "
                "in this case you cannot specify device mesh for each model RPC. "
                "All model RPC will be allocated on GPUs automatically "
                f"allocated according to n_nodes {self.n_nodes} "
                f"and n_gpus_per_node {self.n_gpus_per_node}."
            )

        self.__check_legal_experiment()

        rpcs = self.rpcs
        model_worker = []

        if self.allocation_mode == "search":
            # assert self.mode == "slurm"
            # assumes gradient checkpointing for all training RPCs if one is enabled
            # for the simplicity of search configurations
            rpc_allocs = self._search()
            for rpc_alloc in rpc_allocs:
                assert isinstance(rpc_alloc.rpc, str)
                for rpc in rpcs.values():
                    if rpc.name == rpc_alloc.rpc:
                        rpc_alloc.rpc = rpc
                        break
                else:
                    raise ValueError(f"RPC {rpc_alloc.rpc} not found in rpcs.")
        elif (
            self.allocation_mode == "pipe_data" or self.allocation_mode == "pipe_model"
        ):
            rpc_allocs: List[RPCAllocation] = [
                RPCAllocation(
                    rpc=rpc,
                    device_mesh=self.global_device_mesh,
                    parallel=ParallelismConfig(
                        data_parallel_size=(
                            self.n_gpus_per_node
                            if self.allocation_mode == "pipe_data"
                            else 1
                        ),
                        pipeline_parallel_size=self.n_nodes,
                        model_parallel_size=(
                            self.n_gpus_per_node
                            if self.allocation_mode == "pipe_model"
                            else 1
                        ),
                        use_sequence_parallel=(
                            rpc.interface_type == ModelInterfaceType.TRAIN_STEP
                            and self.allocation_mode == "pipe_model"
                        ),
                    ),
                )
                for rpc in self.rpcs.values()
            ]
        elif self.allocation_mode == "manual":
            rpc_allocs: List[RPCAllocation] = [
                RPCAllocation(
                    rpc=rpc,
                    device_mesh=(
                        make_device_mesh_from_name(
                            self.global_device_mesh.name,
                            self.allocations[rpc_type].device_mesh,
                        )
                        if self.allocations[rpc_type].device_mesh is not None
                        else self.global_device_mesh
                    ),
                    parallel=self.allocations[rpc_type].parallel,
                )
                for rpc_type, rpc in self.rpcs.items()
            ]
        elif self.allocation_mode == "heuristic":
            rpc_allocs: List[RPCAllocation] = self._heuristic_rpc_allocation()
        else:
            raise NotImplementedError()

        shard_counter = defaultdict(lambda: 0)
        resolve_rpc_hooks(rpc_allocs)  # inplace modify MFCDefs in rpc allocations
        import pprint

        pprint.pprint(rpc_allocs)

        model_name_to_rpc_allocs: Dict[ModelName, List[RPCAllocation]] = defaultdict(
            list
        )
        for rpc_alloc in rpc_allocs:
            model_name_to_rpc_allocs[rpc_alloc.rpc.model_name].append(rpc_alloc)

        for i, j in itertools.product(range(self.n_nodes), range(self.n_gpus_per_node)):
            mw = ModelWorker(
                seed=self.seed,
                shards=[],
                datasets=self.datasets,
                cuda_cache_cleanliness=False,
                cuda_cache_clear_freq=10,
                tokenizer_name_or_path=self.tokenizer_name_or_path,
            )
            # print(f"Setting up ModelWorker ({i},{j})")

            for (
                model_name,
                model_rpc_allocs,
            ) in model_name_to_rpc_allocs.items():
                rpcs = [rpc_alloc.rpc for rpc_alloc in model_rpc_allocs]
                rpc_alloc = model_rpc_allocs[0]
                model_cfg = self.models[model_name.role]
                model = make_model_config(model_cfg)
                mapping = rpc_alloc.device_mesh.mapping
                gradient_checkpointing = model_cfg.gradient_checkpointing and any(
                    rpc.interface_type == ModelInterfaceType.TRAIN_STEP for rpc in rpcs
                )

                topo = get_topo(
                    rpc_alloc.parallel,
                    gradient_checkpointing=gradient_checkpointing,
                    max_prompt_len=(
                        self.max_prompt_len
                        if any(
                            rpc.interface_type == ModelInterfaceType.GENERATE
                            for rpc in rpcs
                        )
                        else None
                    ),
                )

                if any(
                    rpc.interface_type == ModelInterfaceType.TRAIN_STEP for rpc in rpcs
                ):
                    backend = make_train_backend_config(model_cfg, rpc_alloc.parallel)
                else:
                    backend = make_inf_backend_config(model_cfg, rpc_alloc.parallel)

                # print(f"model name {model_name}, device mesh name {rpc_alloc.device_mesh.name},"
                #       f"mapping {mapping}")
                if mapping[i, j]:
                    shard_idx = shard_counter[model_name]
                    # print(f"Setting up Shard {shard_idx}, "
                    #       f"(dp, pp, mp) = ({topo.get_coord(shard_idx).data}, "
                    #       f"{topo.get_coord(shard_idx).pipe}, {topo.get_coord(shard_idx).model})")
                    mw.shards.append(
                        StandaloneModelShard(
                            id=ModelShardID(
                                model_name=model_name,
                                topo=topo,
                                dp_rank=topo.get_coord(shard_idx).data,
                                pp_rank=topo.get_coord(shard_idx).pipe,
                                mp_rank=topo.get_coord(shard_idx).model,
                            ),
                            model=model,
                            backend=backend,
                            eval_datasets=self.eval_datasets,
                            eval_dataloader=self.eval_dataloader,
                        )
                    )
                    shard_counter[model_name] += 1
            model_worker.append(mw)

        return ExperimentConfig(
            exp_ctrl=self.exp_ctrl,
            model_rpcs=[rpc_alloc.rpc for rpc_alloc in rpc_allocs],
            model_worker=model_worker,
        )

    def __check_legal_experiment(self):
        if self.n_nodes > 1 and self.mode == "local":
            raise ValueError(
                "Cannot run multi-node experiment in local mode, "
                "please setup slurm for distributed runs."
            )

        if self.n_gpus_per_node != 8 and self.allocation_mode in [
            "search",
            "heuristic",
        ]:
            raise ValueError(
                f"Cannot run search or heuristic allocation with "
                f"n_gpus_per_node {self.n_gpus_per_node}, "
                "please set n_gpus_per_node to 8."
            )

        for rpc_name, rpc in self.rpcs.items():
            if not check_is_realhf_native_model_interface(
                rpc.interface_impl.type_
            ) and self.allocation_mode in ["search", "heuristic"]:
                raise ValueError(
                    f"RPC {rpc.name} interface is not a realhf native implementation. "
                    f"Search and heuristic allocation mode are not available."
                )
            if self.allocation_mode == "manual" and rpc_name not in self.allocations:
                if rpc_name not in self.allocations:
                    raise ValueError(
                        f"RPC {rpc_name} is not in allocations, please implement "
                        f"`allocations()` method in your config class to enable "
                        f"manual allocation."
                    )

            if rpc.model_name.role not in self.models.keys():
                raise ValueError(
                    f"RPC {rpc.name} model name {rpc.model_name.role} is not in models."
                )



from typing import *

import torch
import torch.nn as nn
from torch.nn import init

from realhf.impl.model.parallelism.model_parallel.modules import ParallelEmbedding


class OffsetPositionalEmbedding(nn.Embedding):

    def __init__(
        self,
        num_embeddings: int,
        embedding_dim: int,
        offset: int,
        dtype: Optional[torch.dtype] = None,
        device: Optional[Union[str, torch.device]] = None,
    ):
        # OPT is set up so that if padding_idx is specified then offset the embedding ids by 2
        # and adjust num_embeddings appropriately. Other models don't have this hack
        self.__offset = offset
        super().__init__(
            num_embeddings + self.__offset,
            embedding_dim,
            dtype=dtype,
            device=device,
        )

    def forward(self, position_ids: torch.LongTensor):
        return super().forward(position_ids + self.__offset)


class OffsetParallelPositionalEmbedding(ParallelEmbedding):
    def __init__(
        self,
        num_embeddings: int,
        embedding_dim: int,
        offset: int,
        init_method=init.xavier_normal_,
        # params_dtype: torch.dtype=torch.float32,
        perform_initialization: bool = True,
        dtype: Optional[torch.dtype] = None,
        device: Optional[Union[str, torch.device]] = None,
    ):
        self.__offset = offset
        super(OffsetParallelPositionalEmbedding, self).__init__(
            num_embeddings=num_embeddings + offset,
            embedding_dim=embedding_dim,
            init_method=init_method,
            perform_initialization=perform_initialization,
            dtype=dtype,
            device=device,
        )

    def forward(self, input_: torch.LongTensor) -> torch.Tensor:
        return super().forward(input_ + self.__offset)



import dataclasses
import gc
import os
import time
from contextlib import nullcontext
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import torch
import torch.distributed as dist

import realhf.base.constants as constants
import realhf.base.logging as logging
import realhf.impl.model.parallelism.model_parallel.custom_all_reduce as custom_all_reduce

logger = logging.getLogger("CUDAGraph")

CUDA_GRAPH_STORAGE: Dict[str, torch.cuda.CUDAGraph] = dict()
CUDA_GRAPH_INPUT_BUFFER: Dict[str, Dict[str, torch.Tensor]] = dict()
CUDA_GRAPH_OUTPUT_BUFFER: Dict[str, Dict[str, torch.Tensor]] = dict()


def reinitialize_input_buffer(cuda_graph_name, new_buf):
    global CUDA_GRAPH_INPUT_BUFFER
    assert (
        cuda_graph_name in CUDA_GRAPH_INPUT_BUFFER
    ), f"CUDAGraph {cuda_graph_name} does not exist."

    buf = CUDA_GRAPH_INPUT_BUFFER[cuda_graph_name]
    for k, v in buf.items():
        if torch.is_tensor(v):
            v.copy_(new_buf[k])
        elif isinstance(v, list):
            for i, vv in enumerate(v):
                if torch.is_tensor(vv):
                    vv.copy_(new_buf[k][i])
        else:
            buf[k] = new_buf[k]


@torch.no_grad()
def capture_func(
    name: str,
    func: Callable,
    input_buffer: Dict[str, Any],
    force_recapture: bool = False,
    no_grad: bool = False,
) -> Tuple[torch.cuda.CUDAGraph, Dict[str, torch.Tensor], Dict[str, torch.Tensor]]:
    """Capture a function with cuda graph, store the graph and input/output buffers by name.
    The input/output metadata should match the inputs and outputs of function.

    :param name: The identifier of the CUDAGraph to be captured/reused.
    :type name: str
    :param func: The function to be captured.
    :type func: Callable
    :param input_buffer: The input buffer of the function.
    :type input_buffer: Dict[str, Any]
    :param force_recapture: Whether to force recapture the function.
    :type force_recapture: bool
    :param no_grad: Whether to run the function in no_grad context.
    :type no_grad: bool
    """
    global CUDA_GRAPH_STORAGE
    global CUDA_GRAPH_INPUT_BUFFER
    global CUDA_GRAPH_OUTPUT_BUFFER
    if not force_recapture:
        if name in CUDA_GRAPH_STORAGE:
            reinitialize_input_buffer(name, input_buffer)
            assert name in CUDA_GRAPH_INPUT_BUFFER
            assert name in CUDA_GRAPH_OUTPUT_BUFFER
            return (
                CUDA_GRAPH_STORAGE[name],
                CUDA_GRAPH_INPUT_BUFFER[name],
                CUDA_GRAPH_OUTPUT_BUFFER[name],
            )

    if not custom_all_reduce.is_initialized():
        custom_all_reduce.init_custom_ar()
    assert (
        custom_all_reduce.is_initialized() or constants.model_parallel_world_size() == 1
    )
    assert not constants.sequence_parallel()

    maybe_no_grad = nullcontext() if not no_grad else torch.no_grad()
    st = time.monotonic()

    with custom_all_reduce.graph_capture(), maybe_no_grad:
        logger.info(f"Rank {dist.get_rank()}: Capturing CUDA graph for {name}")
        func(**input_buffer)
        torch.cuda.synchronize()

        graph = torch.cuda.CUDAGraph()
        with torch.cuda.graph(graph):
            output = func(**input_buffer)

        torch.cuda.synchronize()

        logger.info(
            f"Rank {dist.get_rank()}: Capturing CUDA graph {name} "
            f"for decoding takes {time.monotonic() - st:.2f} seconds."
        )

    assert torch.is_tensor(output)
    output_buffer = dict(output=output)

    gc.collect()
    torch.cuda.empty_cache()
    gc.collect()

    CUDA_GRAPH_STORAGE[name] = graph
    CUDA_GRAPH_INPUT_BUFFER[name] = input_buffer
    CUDA_GRAPH_OUTPUT_BUFFER[name] = output_buffer
    return graph, input_buffer, output_buffer


def input_buffer_handle(graph_name: str, tensor_name: str):
    if graph_name not in CUDA_GRAPH_INPUT_BUFFER:
        return None
    if tensor_name not in CUDA_GRAPH_INPUT_BUFFER[graph_name]:
        raise ValueError(
            f"Tensor {tensor_name} not found in input buffer of graph {graph_name}, "
            f"Existing keys = {CUDA_GRAPH_INPUT_BUFFER[graph_name].keys()}"
        )
    return CUDA_GRAPH_INPUT_BUFFER[graph_name][tensor_name]


def output_buffer_handle(graph_name: str, tensor_name: str):
    if graph_name not in CUDA_GRAPH_OUTPUT_BUFFER:
        return None
    if tensor_name not in CUDA_GRAPH_OUTPUT_BUFFER[graph_name]:
        raise ValueError(
            f"Tensor {tensor_name} not found in output buffer of graph {graph_name}, "
            f"existing keys = {CUDA_GRAPH_OUTPUT_BUFFER[graph_name].keys()}"
        )
    return CUDA_GRAPH_OUTPUT_BUFFER[graph_name][tensor_name]


def get_graph(name: str):
    return CUDA_GRAPH_STORAGE.get(name, None)



import dataclasses
import gc
import multiprocessing as mp
import os
import pickle
import queue
import random
import time
import traceback
from typing import Any, Callable, Dict, List, Optional, Tuple

import pynvml
import pytest
import torch
import torch.utils.data

import realhf.api.core.data_api as data_api
from realhf.base import (
    constants,
    gpu_utils,
    logging,
    name_resolve,
    namedarray,
    names,
    topology,
)
from realhf.base.topology import ParallelGrid, PipeModelDataParallelTopology

logger = logging.getLogger("testing")

MODEL_NAME = "default"
_DEFAULT_EXPR_NAME = "test"
_DEFAULT_TRIAL_NAME = "test"


class StandaloneTestingProcess(mp.Process):
    """Aims for defining this class:
    + Removing duplicate setup GPU codes in each test.

    Note that `init_global_constants` should be called in `func`.
    """

    def __init__(
        self,
        rank: int,
        world_size: int,
        barrier: mp.Barrier,
        err_queue: mp.Queue,
        func: Callable,
        *args,
        expr_name: str = None,
        trial_name: str = None,
        **kwargs,
    ):
        super().__init__()
        self.rank = rank
        self.world_size = world_size
        self.err_queue = err_queue
        self.barrier = barrier

        self.expr_name = expr_name if expr_name is not None else _DEFAULT_EXPR_NAME
        self.trial_name = trial_name if trial_name is not None else _DEFAULT_TRIAL_NAME

        self.func = func
        self.args = args
        self.kwargs = kwargs

    def _run(self):
        return self.func(*self.args, **self.kwargs)

    def run(self) -> None:
        assert not torch.cuda.is_initialized()
        torch.cuda.set_device(0)

        self.barrier.wait()

        # init process group
        gpu_utils.reveal_ddp_identity(self.expr_name, self.trial_name, self.rank)
        self.barrier.wait()
        from realhf.impl.model.comm.global_comm import setup_global_comm

        setup_global_comm(self.expr_name, self.trial_name, self.rank)
        # NOTE: The import must be here.
        import deepspeed

        deepspeed.init_distributed()

        # setup some useful constants
        constants.set_experiment_trial_names(self.expr_name, self.trial_name)

        # misc setup
        pynvml.nvmlInit()
        pytorch_memory_burnin(self.rank)

        try:
            self._run()
        except Exception as e:
            print(traceback.format_exc(), flush=True)
            self.err_queue.put(e)
            raise e


class LocalMultiProcessTest:
    """Aims for defining this class:
    1. Defining a barrier and a queue for all sub-processes.
    2. Error handling after launch.
    """

    # NOTE: This is necessary for running pytest, otherwise
    # pytest will exit early before subprocesses terminate.
    mp.set_start_method("spawn", force=True)

    def __init__(
        self,
        world_size: int,
        func: Callable,
        *args,
        expr_name: str = None,
        trial_name: str = None,
        **kwargs,
    ):
        self.barrier = mp.Barrier(world_size)
        self.err_queue = mp.Queue(world_size)
        os.environ["REAL_MODE"] = "LOCAL"
        os.environ["CUDA_DEVICE_MAX_CONNECTIONS"] = "1"
        os.environ["GPU_DEVICES_ISOLATED"] = str(1)
        clear_name_resolve(expr_name, trial_name)
        self.processes = []
        for rank in range(world_size):
            os.environ["CUDA_VISIBLE_DEVICES"] = str(rank)
            p = StandaloneTestingProcess(
                rank,
                world_size,
                self.barrier,
                self.err_queue,
                func,
                *args,
                expr_name=expr_name,
                trial_name=trial_name,
                **kwargs,
            )
            p.start()
            self.processes.append(p)

    def launch(self):
        while any([p.is_alive() for p in self.processes]):
            try:
                err = self.err_queue.get_nowait()
                [p.terminate() for p in self.processes]
                raise err
            except queue.Empty:
                time.sleep(0.1)
        [p.join() for p in self.processes]


def init_global_constants(
    num_dp,
    num_mp,
    num_pp,
    topo=None,
    model_name=None,
    msid2mwid=None,
    sequence_parallel=False,
    gradient_checkpointing=True,
    max_prompt_len=None,
):
    model_name = model_name if model_name is not None else MODEL_NAME

    if topo is None:
        topo = PipeModelDataParallelTopology(
            num_dp=num_dp,
            num_mp=num_mp,
            num_pp=num_pp,
            sequence_parallel=sequence_parallel,
            gradient_checkpointing=gradient_checkpointing,
            max_prompt_len=max_prompt_len,
        )
        ws = num_dp * num_mp * num_pp
    else:
        ws = topo.world_size()

    with constants.model_scope(model_name):
        constants.set_rank_mapping(model_name, topo, msid2mwid=msid2mwid)
        wg_ranks = [constants.to_global_pg_rank(i) for i in range(ws)]
        wg = topology.new_or_get_group(ranks=wg_ranks)

        constants.set_parallelism_group(
            model_name=model_name, pgroup=wg, ranks=wg_ranks
        )
        grid = ParallelGrid(
            process_group=wg,
            topology=topo,
            rank_mapping=constants.rank_mapping_of_model(model_name),
        )
        constants.set_grid(model_name=model_name, grid=grid)


def clear_name_resolve(expr_name=None, trial_name=None):
    expr_name = expr_name if expr_name is not None else _DEFAULT_EXPR_NAME
    trial_name = trial_name if trial_name is not None else _DEFAULT_TRIAL_NAME
    name_resolve.clear_subtree(
        names.trial_root(experiment_name=expr_name, trial_name=trial_name)
    )


def pytorch_memory_burnin(rank):
    torch.cuda.set_device(0)
    torch.cuda.init()
    x = torch.randn(1, device="cuda", dtype=torch.float64, requires_grad=True)
    y = x * torch.randn(1000, device="cuda", dtype=torch.float64)
    y.mean().backward()
    del x, y
    gc.collect()
    torch.cuda.empty_cache()
    gc.collect()


def clear_gpu_cache():
    gc.collect()
    torch.cuda.empty_cache()
    gc.collect()


def get_memory(rank):
    handle = pynvml.nvmlDeviceGetHandleByIndex(rank)
    memory_info = pynvml.nvmlDeviceGetMemoryInfo(handle)
    used_memory = memory_info.used / (1024**2)  # Convert bytes to megabytes
    return used_memory


def get_pytorch_profiler(save_fn: str):

    def trace_handler(p: torch.profiler._KinetoProfile):
        # print(
        #     p.key_averages().table(
        #         sort_by="cuda_memory_usage", row_limit=20, max_name_column_width=30, max_src_column_width=30
        #     )
        # )
        p.export_chrome_trace(save_fn)

    return torch.profiler.profile(
        activities=[
            torch.profiler.ProfilerActivity.CPU,
            torch.profiler.ProfilerActivity.CUDA,
        ],
        record_shapes=True,
        profile_memory=True,
        with_stack=True,
        on_trace_ready=trace_handler,
        with_flops=True,
    )


def random_sample(num_sequences: int, seq_len: int, vocab_size: int, seed: int = 1):
    torch.manual_seed(seed)
    return torch.randint(0, vocab_size, (num_sequences, seq_len), dtype=torch.long)


def make_random_packed_batches(
    n_batches,
    batch_size,
    seq_len,
    vocab_size,
    seed: int = 1,
    dp_rank=None,
    dp_size=None,
):
    assert (dp_rank is None and dp_size is None) or (
        dp_rank is not None and dp_size is not None
    )
    if dp_rank is None:
        dp_rank = constants.data_parallel_rank()
        dp_size = constants.data_parallel_world_size()
    assert batch_size % dp_size == 0
    dp_batch_size = batch_size // dp_size
    n_seqs = batch_size * n_batches
    seqs = random_sample(batch_size * n_batches, seq_len, vocab_size, seed)
    seqs = seqs[n_seqs * dp_rank // dp_size : n_seqs * (dp_rank + 1) // dp_size]
    seqs = [namedarray.NamedArray(packed_prompts=seqs[i]) for i in range(len(seqs))]
    for seq in seqs:
        seq.register_metadata(seqlens=[seq_len])
    batches = [
        seqs[j * dp_batch_size : (j + 1) * dp_batch_size] for j in range(n_batches)
    ]
    batches = [data_api.gather_sequences(batch) for batch in batches]
    return batches


def make_random_unpacked_batches(
    n_batches,
    batch_size,
    seq_len,
    vocab_size,
    seed: int = 1,
    dp_rank=None,
    dp_size=None,
):
    n_seqs = batch_size * n_batches
    dp_batch_size = batch_size // dp_size
    assert (dp_rank is None and dp_size is None) or (
        dp_rank is not None and dp_size is not None
    )
    if dp_rank is None:
        dp_rank = constants.data_parallel_rank()
        dp_size = constants.data_parallel_world_size()
    assert batch_size % dp_size == 0
    seqs = random_sample(batch_size * n_batches, seq_len, vocab_size, seed)
    seqs = seqs[n_seqs * dp_rank // dp_size : n_seqs * (dp_rank + 1) // dp_size]
    batches = [
        seqs[j * dp_batch_size : (j + 1) * dp_batch_size] for j in range(n_batches)
    ]
    batches = [
        dict(
            input_ids=batch,
            attention_mask=torch.ones_like(batch),
        )
        for batch in batches
    ]
    return batches



import dataclasses
import gc
import itertools
import queue
from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Tuple, Union

import torch
import torch.distributed
import torch.utils.checkpoint
import transformers

import realhf.impl.model.utils.cuda_graph as cuda_graph
from realhf.api.core.model_api import GenerationHyperparameters, ReaLModelConfig
from realhf.base import constants, logging
from realhf.impl.model.nn.real_llm_base import PipeCacheData, PipeTransferData
from realhf.impl.model.utils.functional import mask_eos_token
from realhf.impl.model.utils.logits_warper import top_k_top_p_logits
from realhf.impl.model.utils.padding import index_first_axis, unpad_input

if TYPE_CHECKING:
    from .real_llm_api import ReaLModel

logger = logging.getLogger("ReaLModel Generation")


def genstep(
    next_token_logits: torch.Tensor,
    tokenizer: transformers.PreTrainedTokenizerFast,
    unfinished_sequences: torch.Tensor,
    generated_idx: Union[torch.IntTensor, int],
    gconfig: GenerationHyperparameters,
) -> Tuple[torch.Tensor, torch.Tensor, Optional[torch.Tensor], bool, torch.Tensor]:
    """Advance generation by one step given logits.

    Args:
        next_token_logits (torch.Tensor): Shape [bs, vocab_size].
        tokenizer (transformers.PreTrainedTokenizerFast): .
        unfinished_sequences (torch.Tensor): Bool tensor indicator of whether a sequence is finished.
            Shape [bs].
        generated_idx (int): The token index to be generated.
        gconfig (GenerationHyperparameters): .

    Returns:
        Tuple[torch.Tensor, torch.Tensor, torch.Tensor, bool, torch.Tensor]:
        A tuple of
            next_tokens: Shape [bs].
            logprob: The log probability of selected tokens. May be re-normalized
                according to the mask machanism. Shape [bs].
            logits_mask: The mask of logits. Shape [bs, vocab_size].
            terminate: Whether the generation should be terminated.
            unfinished_sequences: Bool tensor indicator of whether a sequence is finished.
                Shape [bs].
    """
    if constants.model_parallel_world_size() > 1:
        from realhf.impl.model.parallelism.model_parallel.mappings import (
            gather_from_tensor_model_parallel_region,
        )

        next_token_logits = gather_from_tensor_model_parallel_region(next_token_logits)

    unfinished_sequences = unfinished_sequences.bool()
    next_token_logits = next_token_logits.float()
    if isinstance(generated_idx, int):
        if generated_idx < gconfig.min_new_tokens:
            next_token_logits = mask_eos_token(
                next_token_logits, eos_token_id=tokenizer.eos_token_id
            )
    else:
        assert isinstance(generated_idx, torch.Tensor)
        if (generated_idx < gconfig.min_new_tokens).any():
            _batch_indices = (generated_idx < gconfig.min_new_tokens).unsqueeze(1)
            _vocab_indices = _batch_indices.new_zeros((1, next_token_logits.shape[1]))
            if tokenizer.eos_token_id is not None:
                _vocab_indices[:, tokenizer.eos_token_id] = 1
            next_token_logits.masked_fill_(
                _batch_indices * _vocab_indices,
                torch.finfo(next_token_logits.dtype).min,
            )

    if not gconfig.greedy:
        next_token_logits /= gconfig.temperature
        next_token_logits = top_k_top_p_logits(
            next_token_logits,
            top_k=gconfig.top_k,
            top_p=gconfig.top_p,
            inplace=True,
            ordered=False,
        )

    distrb = torch.distributions.Categorical(logits=next_token_logits)
    next_tokens = distrb.mode if gconfig.greedy else distrb.sample()
    logprob = distrb.log_prob(next_tokens)

    if constants.model_parallel_world_size() > 1:
        if constants.model_parallel_rank() > 0:
            logprob[:] = 0
            next_tokens[:] = 0
        handle = torch.distributed.all_reduce(
            logprob,
            torch.distributed.ReduceOp.SUM,
            async_op=True,
            group=constants.model_parallel_group(),
        )
        torch.distributed.all_reduce(
            next_tokens,
            torch.distributed.ReduceOp.SUM,
            group=constants.model_parallel_group(),
        )

    if tokenizer.eos_token_id is not None:
        if tokenizer.pad_token_id is None:
            raise ValueError(
                "If `eos_token_id` is defined, make sure that `pad_token_id` is defined."
            )
        next_tokens.masked_fill_(
            unfinished_sequences.logical_not(), tokenizer.pad_token_id
        )
        # next_tokens = next_tokens * unfinished_sequences + tokenizer.pad_token_id * (1 - unfinished_sequences)
    # unfinished_sequences = next_tokens.ne(tokenizer.eos_token_id).long() * unfinished_sequences
    unfinished_sequences.logical_and_(next_tokens.ne(tokenizer.eos_token_id))

    # terminate check
    if isinstance(generated_idx, int):
        terminate = (generated_idx >= gconfig.max_new_tokens - 1) or (
            unfinished_sequences.max() == 0
        )
    else:
        unfinished_sequences.logical_and_(generated_idx < gconfig.max_new_tokens - 1)
        terminate = unfinished_sequences.max() == 0

    logits_mask = next_token_logits == torch.finfo(next_token_logits.dtype).min
    if not logits_mask.any():
        logits_mask = None

    if constants.model_parallel_world_size() > 1:
        handle.wait()

    return next_tokens, logprob, logits_mask, terminate, unfinished_sequences


def prepare_generate_inputs(
    module: "ReaLModel",
    gconfig: GenerationHyperparameters,
    x: PipeTransferData,
    ys: List[PipeCacheData],
    cuda_graph_name: str,
):
    cu_seqlens = x.cu_seqlens
    input_lens = cu_seqlens[1:] - cu_seqlens[:-1]
    # assert constants.pipe_parallel_world_size() >= 2
    layer_indices = range(module.layer_idx_start, module.layer_idx_end)
    min_layer_index = module.layer_idx_start
    bs = input_lens.shape[0]
    cache_seqlens = input_lens.clone().to(dtype=torch.int32)

    block_ys = ys
    assert len(layer_indices) == len(block_ys), (len(block_ys), layer_indices)
    if constants.is_first_pipe_stage():
        layer_indices = layer_indices[1:]
        ys[0].cache_seqlens = cache_seqlens
        block_ys = ys[1:]
    if constants.is_last_pipe_stage():
        layer_indices = layer_indices[:-1]
        ys[-1].cache_seqlens = cache_seqlens
        block_ys = block_ys[:-1]

    for y, layer_idx in zip(block_ys, layer_indices):
        assert (
            y.k_cache is not None
            and y.v_cache is not None
            and y.cache_seqlens is not None
        ), (y.k_cache is None, y.v_cache is None, y.cache_seqlens is None)
        kvcache_seqlen = max(
            constants.max_prompt_len() + gconfig.max_new_tokens,
            module.config.hidden_dim // module.config.head_dim + 10,
        )
        k_cache_handle = cuda_graph.input_buffer_handle(cuda_graph_name, "k_caches")
        v_cache_handle = cuda_graph.input_buffer_handle(cuda_graph_name, "v_caches")
        if k_cache_handle is not None and v_cache_handle is not None:
            k_cache = k_cache_handle[layer_idx - min_layer_index][:bs, :kvcache_seqlen]
            v_cache = v_cache_handle[layer_idx - min_layer_index][:bs, :kvcache_seqlen]
        else:
            k_cache = torch.zeros(
                (bs, kvcache_seqlen, *y.k_cache.shape[1:]),
                dtype=y.k_cache.dtype,
                device=y.k_cache.device,
            )
            v_cache = torch.zeros(
                (bs, kvcache_seqlen, *y.v_cache.shape[1:]),
                dtype=y.v_cache.dtype,
                device=y.v_cache.device,
            )
        indices = (
            torch.arange(kvcache_seqlen, device=module.device, dtype=torch.long)[
                None, :
            ]
            < input_lens[:, None]
        )
        k_cache[indices] = y.k_cache
        v_cache[indices] = y.v_cache
        y.k_cache = k_cache
        y.v_cache = v_cache
        y.cache_seqlens = cache_seqlens

    return x, ys


def maybe_capture_cudagraph(
    module: "ReaLModel",
    x: PipeTransferData,
    ys: List[PipeCacheData],
    cuda_graph_name: str,
):
    bs = x.cu_seqlens.shape[0] - 1
    cache_seqlens = ys[0].cache_seqlens
    input_buffers = dict(
        input_ids=torch.ones((bs,), dtype=torch.long, device="cuda"),
        position_ids=cache_seqlens.clone(),
        k_caches=[y.k_cache for y in ys],
        v_caches=[y.v_cache for y in ys],
        cache_seqlens=cache_seqlens.clone(),
        max_seqlen=x.max_seqlen,
        cu_seqlens=x.cu_seqlens.clone(),
        hidden_states=(
            torch.zeros(
                (bs, x.pp_input.shape[1]), dtype=x.pp_input.dtype, device="cuda"
            )
            if x.pp_input is not None
            else None
        ),
    )

    graph, input_buffers, output_buffers = cuda_graph.capture_func(
        cuda_graph_name,
        module._forward,
        input_buffers,
        force_recapture=False,
        no_grad=True,
    )

    return graph, input_buffers, output_buffers


@torch.no_grad()
def generate(
    model: "ReaLModel",
    tokenizer: transformers.PreTrainedTokenizerFast,
    packed_input_ids: Optional[torch.LongTensor] = None,
    cu_seqlens: Optional[torch.LongTensor] = None,
    max_seqlen: Optional[int] = None,
    gconfig: GenerationHyperparameters = dataclasses.field(
        default_factory=GenerationHyperparameters
    ),
) -> Tuple[
    torch.Tensor,
    torch.Tensor,
    torch.Tensor,
    List[PipeCacheData],
    Optional[torch.Tensor],
]:
    """Generete a sequence with a ReaLModel."""
    bs = cu_seqlens.shape[0] - 1
    device = model.device
    mconfig: ReaLModelConfig = model.config

    terminate = False
    generated_idx = 0
    unfinished_sequences = torch.ones(bs, dtype=torch.long, device=device)

    gen_token_ph = []
    gen_logprob_ph = []
    gen_logits_mask_ph = []

    prompt_logits = None
    # Prepare inputs for generation iterations

    # Input_ids may have different lengths, we should first pack them into a large batch
    # to use varlen flash attention, then record kv caches for the following inferences.
    max_seqlen = (cu_seqlens[1:] - cu_seqlens[:-1]).max().item()
    if constants.max_prompt_len() < max_seqlen:
        raise RuntimeError(
            f"Input sequence length {max_seqlen} is larger than the maximum sequence length "
            f"supported by the model {constants.max_prompt_len()}."
        )
    x = PipeTransferData(
        cu_seqlens=cu_seqlens, max_seqlen=max_seqlen, store_kv_cache=True
    )
    # one embedding layer, n_layers transformer block, one output layer
    ys = [PipeCacheData(packed_input_ids=packed_input_ids)] + [
        PipeCacheData() for _ in range(mconfig.n_layers + 1)
    ]

    # Model forward will set k/v cache in PipeCacheData.
    prompt_logits = model(x, ys)[0].pp_output
    logits = prompt_logits[cu_seqlens[1:] - 1]

    # Next, we will generate the next token after prompts.
    # cache_seqlens is exactly the lengths of prompts.
    # We perform a genstep outside the loop due to a historical reason.
    next_tokens, logprob, logits_mask, terminate, unfinished_sequences = genstep(
        logits, tokenizer, unfinished_sequences, generated_idx, gconfig
    )
    gen_token_ph.append(next_tokens)
    gen_logprob_ph.append(logprob)
    gen_logits_mask_ph.append(logits_mask)
    generated_idx += 1

    cuda_graph_name = "decoding"
    x, ys = prepare_generate_inputs(model, gconfig, x, ys, cuda_graph_name)
    graph = None
    if gconfig.use_cuda_graph:
        graph, input_buffers, output_buffers = maybe_capture_cudagraph(
            model, x, ys, cuda_graph_name
        )

    # The main loop.
    while not terminate:
        # the next round of inference
        if graph is not None:
            input_buffers["input_ids"][:bs].copy_(next_tokens, non_blocking=True)
            input_buffers["position_ids"][:bs].copy_(
                ys[0].cache_seqlens, non_blocking=True
            )
            input_buffers["cache_seqlens"][:bs].copy_(
                ys[0].cache_seqlens, non_blocking=True
            )
            # K/v cache will be changed in-place with flash attention.
            graph.replay()
            logits = output_buffers["output"][:bs].squeeze(1)
        else:
            ys[0].packed_input_ids = next_tokens
            ys[0].packed_position_ids = None
            x.cu_seqlens = torch.arange(bs + 1, dtype=torch.int32, device=device)
            x.max_seqlen = 1
            # K/v cache will be changed in-place with flash attention.
            logits = model(x, ys)[0].pp_output.squeeze(dim=1)

        ys[
            0
        ].cache_seqlens += (
            1  # The global handle. This will increase all handles in ys by 1.
        )
        next_tokens, logprob, logits_mask, terminate, unfinished_sequences = genstep(
            logits, tokenizer, unfinished_sequences, generated_idx, gconfig
        )
        gen_token_ph.append(next_tokens)
        gen_logprob_ph.append(logprob)
        gen_logits_mask_ph.append(logits_mask)
        generated_idx += 1

    gen_tokens, log_probs, logits_mask = _gather_gen_output_from_list(
        gen_token_ph, gen_logprob_ph, gen_logits_mask_ph
    )
    return gen_tokens, log_probs, logits_mask, ys[1:-1], prompt_logits


def _gather_gen_output_from_list(
    gen_token_ph: List[torch.LongTensor],
    gen_logprob_ph: List[torch.FloatTensor],
    gen_logits_mask_ph: List[torch.BoolTensor],
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """Stack over the sequence dimension given a list of single-token tensors."""
    gen_tokens = torch.stack(gen_token_ph, -1)  # [bs, seqlen]
    log_probs = torch.stack(gen_logprob_ph, -1)  # [bs, seqlen]
    if all([m is None for m in gen_logits_mask_ph]):
        logits_mask = None
    else:
        mm = next(m for m in gen_logits_mask_ph if m is not None)
        gen_logits_mask_ph = [
            torch.ones_like(mm) if m is None else m for m in gen_logits_mask_ph
        ]
        logits_mask = torch.stack(gen_logits_mask_ph, 1)  # [bs, seqlen, vocab_size]
    return gen_tokens, log_probs, logits_mask


def _gather_minibatch_gen_outputs(
    all_gen_tokens: List[torch.LongTensor],
    all_log_probs: List[torch.FloatTensor],
    all_logits_mask: List[torch.BoolTensor],
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """Concate over the batch dimension given multiple [bs, seqlen] mini-batch tensors.

    Since different minibatches may have different generated lengths,
    we should pad them to the same length.
    """
    gen_tokens_lengths = [t.shape[-1] for t in all_gen_tokens]
    max_gen_tokens_length = max(gen_tokens_lengths)

    padded_gen_tokens = []
    padded_log_probs = []
    padded_logits_mask = []

    n_mbs = len(all_gen_tokens)
    for i in range(n_mbs):
        assert all_gen_tokens[i].shape == all_log_probs[i].shape
        gen_len = all_gen_tokens[i].shape[-1]

        gen_token = all_gen_tokens[i]
        log_probs = all_log_probs[i]
        logits_mask = all_logits_mask[i]

        if gen_len < max_gen_tokens_length:
            pad_size = max_gen_tokens_length - gen_len
            gen_token = torch.nn.functional.pad(gen_token, (0, pad_size))
            log_probs = torch.nn.functional.pad(log_probs, (0, pad_size))
            if logits_mask is not None:
                logits_mask = torch.nn.functional.pad(
                    logits_mask,
                    (0, 0, 0, pad_size),
                    mode="constant",
                    value=1,
                )

        padded_gen_tokens.append(gen_token)
        padded_log_probs.append(log_probs)
        padded_logits_mask.append(logits_mask)

    gen_tokens = torch.cat(padded_gen_tokens, 0)
    log_probs = torch.cat(padded_log_probs, 0)
    if all([m is None for m in padded_logits_mask]):
        logits_mask = None
    else:
        mm = next(m for m in padded_logits_mask if m is not None)
        padded_logits_mask = [
            torch.ones_like(mm) if m is None else m for m in padded_logits_mask
        ]
        logits_mask = torch.cat(padded_logits_mask, 0)  # [bs, seqlen, vocab_size]

    return (gen_tokens, log_probs, logits_mask)


def concat_prompt_to_generation_output(
    packed_prompts: torch.LongTensor,
    prompt_lengths: torch.IntTensor,
    gen_tokens: torch.LongTensor,
    logprobs: torch.FloatTensor,
    logits_mask: torch.BoolTensor,
    gen_lengths: torch.IntTensor,
) -> Tuple[
    torch.LongTensor,
    torch.FloatTensor,
    torch.BoolTensor,
    torch.IntTensor,
    torch.BoolTensor,
]:
    device = packed_prompts.device

    prompts_list, prompt_log_probs_list, prompt_logits_mask_list = [], [], []
    gen_tokens_list, gen_log_probs_list, gen_logits_mask_list = [], [], []

    bs = prompt_lengths.shape[0]
    prompt_cu_seqlens = torch.nn.functional.pad(prompt_lengths.cumsum(0), (1, 0))
    for i in range(bs):
        prompt_len, gen_len = prompt_lengths[i].item(), gen_lengths[i].item()

        # log_probs is one-step shorter than token sequences.
        prompts_list.append(
            packed_prompts[prompt_cu_seqlens[i] : prompt_cu_seqlens[i + 1]]
        )
        prompt_log_probs_list.append(logprobs.new_zeros(prompt_len - 1))
        if logits_mask is not None:
            prompt_logits_mask_list.append(
                logits_mask.new_ones((prompt_len - 1, logits_mask.shape[-1]))
            )

        # Generated tokens are right-padded.
        gen_tokens_list.append(gen_tokens[i, :gen_len])
        gen_log_probs_list.append(logprobs[i, :gen_len])
        if logits_mask is not None:
            gen_logits_mask_list.append(
                torch.cat(
                    [
                        logits_mask[i, :gen_len],
                        logits_mask.new_ones(1, logits_mask.shape[-1]),
                    ]
                )
            )

    seq = torch.cat(
        list(itertools.chain.from_iterable(zip(prompts_list, gen_tokens_list)))
    )
    seq_lengths = prompt_lengths + gen_lengths
    packed_logprobs = torch.cat(
        list(
            itertools.chain.from_iterable(
                zip(prompt_log_probs_list, gen_log_probs_list)
            )
        )
    )
    assert seq.shape[0] == packed_logprobs.shape[0] + bs, (
        seq.shape,
        packed_logprobs.shape,
        bs,
    )
    packed_logits_mask = None
    if gen_logits_mask_list:
        packed_logits_mask = torch.cat(
            list(
                itertools.chain.from_iterable(
                    zip(prompt_logits_mask_list, gen_logits_mask_list)
                )
            )
        )

    prompt_mask = zip(
        [torch.ones(plen, dtype=torch.bool, device=device) for plen in prompt_lengths],
        [torch.zeros(glen, dtype=torch.bool, device=device) for glen in gen_lengths],
    )
    prompt_mask = torch.cat(list(itertools.chain.from_iterable(prompt_mask)))

    return (seq, packed_logprobs, packed_logits_mask, seq_lengths, prompt_mask)


@torch.no_grad()
def vanilla_packed_generate(
    model: "ReaLModel",
    tokenizer: transformers.PreTrainedTokenizerFast,
    input_ids: torch.Tensor,
    attention_mask: Optional[torch.Tensor] = None,
    gconfig: GenerationHyperparameters = dataclasses.field(
        default_factory=GenerationHyperparameters
    ),
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """Only used for debugging."""
    mconfig: ReaLModelConfig = model.config

    terminate = False
    generated_idx = 0
    unfinished_sequences = torch.ones(
        input_ids.shape[0], dtype=torch.long, device=input_ids.device
    )

    gen_token_ph = []
    gen_logprob_ph = []
    gen_logits_mask_ph = []

    # The main loop.
    while not terminate:
        packed_input_ids, _, cu_seqlens, max_seqlen = unpad_input(
            input_ids, attention_mask
        )
        x = PipeTransferData(cu_seqlens=cu_seqlens, max_seqlen=max_seqlen)
        # one embedding layer, n_layers transformer block, one output layer
        ys = [PipeCacheData(packed_input_ids=packed_input_ids)] + [
            PipeCacheData() for _ in range(mconfig.n_layers + 1)
        ]
        # Model forward will set k/v cache in PipeCacheData.
        logits = model(x, ys).pp_output
        logits = logits[cu_seqlens[1:] - 1]
        # Next, we will generate the next token after prompts.
        # cache_seqlens is exactly the lengths of prompts.
        next_tokens, logprob, logits_mask, terminate, unfinished_sequences = genstep(
            logits, tokenizer, unfinished_sequences, generated_idx, gconfig
        )
        gen_token_ph.append(next_tokens)
        gen_logprob_ph.append(logprob)
        gen_logits_mask_ph.append(logits_mask)
        generated_idx += 1

        input_ids = torch.cat([input_ids, next_tokens.unsqueeze(-1)], 1)
        am = torch.logical_and(
            next_tokens.unsqueeze(-1).not_equal(tokenizer.eos_token_id),
            next_tokens.unsqueeze(-1).not_equal(tokenizer.pad_token_id),
        )
        attention_mask = torch.cat([attention_mask, am], 1)

    gen_tokens = torch.stack(gen_token_ph, -1)
    log_probs = torch.stack(gen_logprob_ph, -1)
    if all([m is None for m in gen_logits_mask_ph]):
        logits_mask = None
    else:
        mm = next(m for m in gen_logits_mask_ph if m is not None)
        gen_logits_mask_ph = [
            torch.ones_like(mm) if m is None else m for m in gen_logits_mask_ph
        ]
        logits_mask = torch.stack(gen_logits_mask_ph, -2)

    return gen_tokens, log_probs, logits_mask


@torch.no_grad()
def vanilla_cpu_generate(
    model: "ReaLModel",
    tokenizer: transformers.PreTrainedTokenizerFast,
    input_ids: torch.Tensor,
    attention_mask: Optional[torch.Tensor] = None,
    gconfig: GenerationHyperparameters = dataclasses.field(
        default_factory=GenerationHyperparameters
    ),
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """Only used for debugging."""
    mconfig: ReaLModelConfig = model.config
    assert str(input_ids.device) == "cpu"

    terminate = False
    generated_idx = 0
    unfinished_sequences = torch.ones(
        input_ids.shape[0], dtype=torch.long, device=input_ids.device
    )

    gen_token_ph = []
    gen_logprob_ph = []
    gen_logits_mask_ph = []

    # The main loop.
    while not terminate:
        x = PipeTransferData(attention_mask=attention_mask)
        # one embedding layer, n_layers transformer block, one output layer
        ys = [PipeCacheData(packed_input_ids=input_ids)] + [
            PipeCacheData() for _ in range(mconfig.n_layers + 1)
        ]
        # Model forward will set k/v cache in PipeCacheData.
        logits = model(x, ys).pp_output[:, -1, :]
        # Next, we will generate the next token after prompts.
        # cache_seqlens is exactly the lengths of prompts.
        next_tokens, logprob, logits_mask, terminate, unfinished_sequences = genstep(
            logits, tokenizer, unfinished_sequences, generated_idx, gconfig
        )
        gen_token_ph.append(next_tokens)
        gen_logprob_ph.append(logprob)
        gen_logits_mask_ph.append(logits_mask)
        generated_idx += 1

        input_ids = torch.cat([input_ids, next_tokens.unsqueeze(-1)], 1)
        am = torch.logical_and(
            next_tokens.unsqueeze(-1).not_equal(tokenizer.eos_token_id),
            next_tokens.unsqueeze(-1).not_equal(tokenizer.pad_token_id),
        )
        attention_mask = torch.cat([attention_mask, am], 1)

    gen_tokens = torch.stack(gen_token_ph, -1)
    log_probs = torch.stack(gen_logprob_ph, -1)
    if all([m is None for m in gen_logits_mask_ph]):
        logits_mask = None
    else:
        mm = next(m for m in gen_logits_mask_ph if m is not None)
        gen_logits_mask_ph = [
            torch.ones_like(mm) if m is None else m for m in gen_logits_mask_ph
        ]
        logits_mask = torch.stack(gen_logits_mask_ph, -2)

    return gen_tokens, log_probs, logits_mask


class InflightBatchingGenerator:

    def __init__(
        self,
        inqueue: queue.Queue,
        outqueue: queue.Queue,
        model: "ReaLModel",
        tokenizer: transformers.PreTrainedTokenizerFast,
        gconfig: GenerationHyperparameters,
        batch_size: int,
        max_prompt_len: int,
    ):
        self.inqueue = inqueue
        self.outqueue = outqueue

        self.model = model
        self.mconfig = mconfig = model.config
        self.tokenizer = tokenizer

        self.gconfig = gconfig
        self.batch_size = batch_size
        self.max_prompt_len = max_prompt_len

        kvcache_seqlen = max(
            max_prompt_len + gconfig.max_new_tokens,
            mconfig.hidden_dim // mconfig.head_dim + 10,
        )
        _p = next(self.model.parameters())
        dtype, device = _p.dtype, _p.device

        # Cache
        self.k_caches = [
            torch.zeros(
                (
                    batch_size,
                    kvcache_seqlen,
                    mconfig.n_kv_heads,
                    mconfig.head_dim,
                ),
                dtype=dtype,
                device=device,
            )
            for _ in range(self.mconfig.n_layers)
        ]
        self.v_caches = [
            torch.zeros(
                (
                    batch_size,
                    kvcache_seqlen,
                    mconfig.n_kv_heads,
                    mconfig.head_dim,
                ),
                dtype=dtype,
                device=device,
            )
            for _ in range(self.mconfig.n_layers)
        ]
        self.cache_seqlens = torch.zeros(
            (batch_size,), dtype=torch.int32, device=device
        )

        # Input buffers
        self.input_buf = torch.zeros(
            (batch_size, max_prompt_len), dtype=torch.long, device=device
        )
        self.input_buf_lens = torch.zeros(
            (batch_size,), dtype=torch.int32, device=device
        )

        # Save prompts for output
        self.prompt_tokens = [None for _ in range(batch_size)]

        # Generation state
        self.generate_idx = torch.zeros((batch_size,), dtype=torch.int32, device=device)
        self.unfinished_sequences = torch.zeros(
            (batch_size,), dtype=torch.float32, device=device
        )

        self.ys = (
            [
                PipeCacheData(
                    cache_seqlens=self.cache_seqlens,
                )
            ]
            + [
                PipeCacheData(k_cache=k, v_cache=v, cache_seqlens=self.cache_seqlens)
                for k, v in zip(self.k_caches, self.v_caches)
            ]
            + [PipeCacheData()]
        )

        # output buffers
        self.output_tokens_buf = [[] for _ in range(batch_size)]
        self.output_logprob_buf = [[] for _ in range(batch_size)]
        self.output_logits_mask = [[] for _ in range(batch_size)]

    def _get_non_eos_logits(self) -> torch.FloatTensor:
        self.ys[0].packed_position_ids = None
        self.ys[0].packed_input_ids = self.input_buf[:, :1]
        logits = self.model(PipeTransferData(), self.ys).pp_output.squeeze(dim=1)

        self.cache_seqlens += 1
        return logits.float()

    def _get_inflight_logits(self) -> torch.FloatTensor:
        finished_sequences = self.unfinished_sequences.logical_not()
        assert finished_sequences.any()

        finish_indices = finished_sequences.nonzero().squeeze(-1).tolist()

        # pop out finished sequences and clear corresponding buffers
        for i in finish_indices:
            prompt_tokens = self.prompt_tokens[i]

            # Used to skip the first call.
            if prompt_tokens is not None:
                gen_tokens = torch.stack(self.output_tokens_buf[i])
                gen_logp = torch.stack(self.output_logprob_buf[i])
                if all([m is None for m in self.output_logits_mask[i]]):
                    gen_logits_mask = None
                else:
                    mm = next(m for m in self.output_logits_mask[i] if m is not None)
                    gen_logits_mask = [
                        torch.ones_like(mm) if m is None else m
                        for m in self.output_logits_mask[i]
                    ]
                    gen_logits_mask = torch.stack(gen_logits_mask, -2)

                res = dict(
                    prompt=prompt_tokens,
                    gen=gen_tokens,
                    logp=gen_logp,
                    logits_mask=gen_logits_mask,
                )
                try:
                    self.outqueue.put_nowait(res)
                except queue.Full as e:
                    raise RuntimeError(
                        "Output queue is full. Please set a larger queue size."
                    ) from e

            # clear cache
            self.cache_seqlens[i] = 0

            # clear input buffers and prompts
            self.input_buf[i] = 0
            self.input_buf_lens[i] = 0
            self.prompt_tokens[i] = None

            # clear generation state
            self.generate_idx[i] = 0
            self.unfinished_sequences[i] = 1

            self.output_logits_mask[i] = []
            self.output_tokens_buf[i] = []
            self.output_logprob_buf[i] = []

        # build packed input ids with variable lengths for the next-step inference
        for i in range(self.batch_size):
            if i in finish_indices:
                try:
                    prompt = self.inqueue.get_nowait()
                    self.prompt_tokens[i] = prompt
                    self.input_buf[i, : prompt.shape[0]] = prompt
                    self.input_buf_lens[i] = prompt.shape[0]
                except queue.Empty as e:
                    raise RuntimeError(
                        "Input queue is empty. This should not happen."
                    ) from e

        input_lens = self.input_buf_lens
        valid_input_mask = torch.arange(
            self.max_prompt_len, device=self.input_buf.device, dtype=torch.int32
        ).unsqueeze(0) < input_lens.unsqueeze(-1)
        indices = torch.nonzero(valid_input_mask.flatten(), as_tuple=False).flatten()
        packed_input_ids = self.input_buf.flatten()[indices]
        max_seqlen = int(max(input_lens))
        cu_seqlens = torch.nn.functional.pad(
            input_lens.cumsum(0), (1, 0), value=0
        ).int()

        x = PipeTransferData(cu_seqlens=cu_seqlens, max_seqlen=max_seqlen)
        self.ys[0].packed_position_ids = None
        self.ys[0].packed_input_ids = packed_input_ids
        logits = self.model(x, self.ys).pp_output
        logits = index_first_axis(logits, (cu_seqlens[1:] - 1).long())

        self.cache_seqlens += input_lens

        return logits.float()

    def advance_one_genstep(self):
        if self.unfinished_sequences.logical_not().any():
            logits = self._get_inflight_logits()
        else:
            logits = self._get_non_eos_logits()

        next_tokens, logprob, logits_mask, _, self.unfinished_sequences = genstep(
            logits,
            self.tokenizer,
            self.unfinished_sequences,
            self.generate_idx,
            self.gconfig,
        )

        for i in range(self.batch_size):
            self.output_tokens_buf[i].append(next_tokens[i].long())
            self.output_logprob_buf[i].append(logprob[i].float())
            if logits_mask is not None:
                self.output_logits_mask[i].append(logits_mask[i].bool())
            else:
                self.output_logits_mask[i].append(None)

        self.generate_idx += 1
        self.input_buf[:, 0] = next_tokens

    def step_for(self, n: int):
        for _ in range(n):
            self.advance_one_genstep()



import os
import random
import time
import uuid
from typing import *

import numpy as np
import pytest
import torch

from realhf.impl.model.nn.flatten_param import (
    _set_intervals_py,
    _slice_intervals_py,
    set_intervals,
    slice_intervals,
)


def make_intervals(maxsize, n_intervals):
    assert maxsize // n_intervals > 1
    s = maxsize // n_intervals
    intervals = []
    interval_size = 0
    for i in range(n_intervals):
        intervals.append((i * s, i * s + s // 2))
        interval_size += s // 2
    return np.array(intervals, dtype=np.int64), interval_size


def maybe_synchronize_cuda():
    if torch.cuda.is_available():
        torch.cuda.synchronize()


@pytest.mark.parametrize("device", [torch.device("cuda"), torch.device("cpu")])
@pytest.mark.parametrize("n_intervals", list(reversed([1, 100, 10000, 100000])))
@pytest.mark.parametrize("dtype", [torch.bfloat16, torch.float32, torch.float16])
def test_get(n_intervals: int, dtype: torch.dtype, device: torch.device):
    if device == torch.device("cuda") and not torch.cuda.is_available():
        return

    input_tensor = torch.randn(int(1e8), device=device, dtype=dtype)
    intervals, _ = make_intervals(input_tensor.size(0), n_intervals)

    # warmup
    slice_intervals(input_tensor, intervals)
    _slice_intervals_py(input_tensor, intervals)

    maybe_synchronize_cuda()
    tik = time.perf_counter()
    for _ in range(10):
        output_tensor = slice_intervals(input_tensor, intervals)
    maybe_synchronize_cuda()
    t1 = time.perf_counter() - tik

    maybe_synchronize_cuda()
    tik = time.perf_counter()
    for _ in range(10):
        o2 = _slice_intervals_py(input_tensor, intervals)
    maybe_synchronize_cuda()
    t2 = time.perf_counter() - tik
    assert torch.allclose(output_tensor, o2)
    print(f"Success! C++ ext time: {t1:.4f}, PyTorch time: {t2:.4f}")


@pytest.mark.skipif(not torch.cuda.is_available(), reason="This test requires a GPU.")
@pytest.mark.parametrize(
    "n_intervals", list(reversed([1, 10, 100, 1000, 2048, 4096, 10000, 100000]))
)
@pytest.mark.parametrize("dtype", [torch.bfloat16, torch.float16, torch.float32])
def test_set(n_intervals: int, dtype: torch.dtype):
    # NOTE: Since the set_intervals degenerate to the python implementation with CPU tensors,
    # We don't need to test it with CPU tensors.

    x = torch.randn(int(1e8), device="cuda", dtype=dtype)
    intervals, interval_size = make_intervals(x.size(0), n_intervals)
    src = torch.randn(interval_size, device="cuda", dtype=dtype)

    # warmup
    input_tensor1 = x.clone()
    set_intervals(src, input_tensor1, intervals)
    input_tensor2 = x.clone()
    _set_intervals_py(src, input_tensor2, intervals)

    input_tensor1 = x.clone()
    maybe_synchronize_cuda()
    tik = time.perf_counter()
    for _ in range(10):
        set_intervals(src, input_tensor1, intervals)
    maybe_synchronize_cuda()
    t1 = time.perf_counter() - tik

    input_tensor2 = x.clone()
    maybe_synchronize_cuda()
    tik = time.perf_counter()
    for _ in range(10):
        _set_intervals_py(src, input_tensor2, intervals)
    maybe_synchronize_cuda()
    t2 = time.perf_counter() - tik

    assert torch.allclose(input_tensor1, input_tensor2)
    print(f"Success! C++ ext time: {t1:.4f}, PyTorch time: {t2:.4f}")



import dataclasses
import math
from typing import *

import torch
import torch.distributed as dist

try:
    from megatron.core.distributed.distributed_data_parallel import (
        DistributedDataParallel as MegatronDDP,
    )
    from megatron.core.optimizer.clip_grads import clip_grad_norm_fp32, count_zeros_fp32
    from megatron.core.optimizer.distrib_optimizer import (
        DistributedOptimizer as MegatronDistOptim,
    )
    from megatron.core.optimizer.distrib_optimizer import (
        MixedPrecisionOptimizer as MegatronMixedPrecOptim,
    )
except (ModuleNotFoundError, ImportError):
    # importing megatron.core in CPU container will fail due to the requirement of apex
    # Here class types must be defined for type hinting
    class MegatronDDP:
        pass

    class MegatronDistOptim:
        pass


from realhf.base import constants, logging

if TYPE_CHECKING:
    from realhf.impl.model.nn.real_llm_api import ReaLModel, ReaLModelBlock

logger = logging.getLogger("Model Backend Utils")


@dataclasses.dataclass
class MegatronEngine:
    ddp: MegatronDDP
    optim: MegatronDistOptim
    lr_scheduler: Any

    def zero_grad(self, set_to_none=True):
        self.ddp.zero_grad_buffer()
        self.optim.zero_grad(set_to_none=set_to_none)


# Adopted from Megatron-LM/megatron/training/optimizer_param_scheduler.py
class OptimizerParamScheduler(object):
    """Anneals learning rate and weight decay.

    Adopted from Megatron-LM. This class is not included in megatron.core,
    so we have to copy-paste it here.
    """

    def __init__(
        self,
        optimizer,
        init_lr,
        max_lr,
        min_lr,
        lr_warmup_steps,
        lr_decay_steps,
        lr_decay_style,
        start_wd,
        end_wd,
        wd_incr_steps,
        wd_incr_style,
        use_checkpoint_opt_param_scheduler=True,
        override_opt_param_scheduler=False,
    ):

        # Class values.
        self.optimizer = optimizer

        self.init_lr = init_lr
        self.max_lr = float(max_lr)
        self.min_lr = min_lr
        assert self.min_lr >= 0.0
        assert self.max_lr >= self.min_lr
        assert self.init_lr <= self.max_lr

        self.lr_warmup_steps = lr_warmup_steps
        self.num_steps = 0
        self.lr_decay_steps = lr_decay_steps
        assert self.lr_decay_steps > 0
        assert self.lr_warmup_steps < self.lr_decay_steps

        self.lr_decay_style = lr_decay_style

        self.start_wd = start_wd
        self.end_wd = end_wd
        assert self.start_wd >= 0.0
        assert self.end_wd >= self.start_wd
        self.wd_incr_steps = wd_incr_steps
        self.wd_incr_style = wd_incr_style

        self.override_opt_param_scheduler = override_opt_param_scheduler
        self.use_checkpoint_opt_param_scheduler = use_checkpoint_opt_param_scheduler
        if self.override_opt_param_scheduler:
            assert not self.use_checkpoint_opt_param_scheduler, (
                "both override and " "use-checkpoint are set."
            )

        # Set the learning rate
        self.step(0)
        self.log_rank_0("> learning rate decay style: {}".format(self.lr_decay_style))

    def log_rank_0(self, msg):
        if constants.parallelism_rank() == 0:
            logger.info(msg)

    def get_wd(self):
        """Weight decay incr functions"""
        if self.num_steps > self.wd_incr_steps:
            return self.end_wd

        if self.wd_incr_style == "constant":
            assert self.start_wd == self.end_wd
            return self.end_wd

        incr_ratio = float(self.num_steps) / float(self.wd_incr_steps)
        assert incr_ratio >= 0.0
        assert incr_ratio <= 1.0
        delta_wd = self.end_wd - self.start_wd

        if self.wd_incr_style == "linear":
            coeff = incr_ratio
        elif self.wd_incr_style == "cosine":
            coeff = 0.5 * (math.cos(math.pi * (1 - incr_ratio)) + 1.0)
        else:
            raise Exception(
                "{} weight decay increment style is not supported.".format(
                    self.wd_incr_style
                )
            )

        return self.start_wd + coeff * delta_wd

    def get_lr(self, param_group):
        """Learning rate decay functions from:
        https://openreview.net/pdf?id=BJYwwY9ll pg. 4"""

        max_lr = param_group.get("max_lr", self.max_lr)
        min_lr = param_group.get("min_lr", self.min_lr)

        # Use linear warmup for the initial part.
        if self.lr_warmup_steps > 0 and self.num_steps <= self.lr_warmup_steps:
            return self.init_lr + (
                (max_lr - self.init_lr)
                * float(self.num_steps)
                / float(self.lr_warmup_steps)
            )

        # If the learning rate is constant, just return the initial value.
        if self.lr_decay_style == "constant":
            return max_lr

        # For any steps larger than `self.lr_decay_steps`, use `min_lr`.
        if self.num_steps > self.lr_decay_steps:
            return min_lr

        # If we are done with the warmup period, use the decay style.
        if self.lr_decay_style == "inverse-square-root":
            warmup_steps = max(self.lr_warmup_steps, 1)
            num_steps = max(self.num_steps, 1)
            lr = max_lr * warmup_steps**0.5 / (num_steps**0.5)
            return max(min_lr, lr)

        num_steps_ = self.num_steps - self.lr_warmup_steps
        decay_steps_ = self.lr_decay_steps - self.lr_warmup_steps
        decay_ratio = float(num_steps_) / float(decay_steps_)
        assert decay_ratio >= 0.0
        assert decay_ratio <= 1.0
        delta_lr = max_lr - min_lr

        if self.lr_decay_style == "linear":
            coeff = 1.0 - decay_ratio
        elif self.lr_decay_style == "cosine":
            coeff = 0.5 * (math.cos(math.pi * decay_ratio) + 1.0)
        else:
            raise Exception(
                "{} decay style is not supported.".format(self.lr_decay_style)
            )

        return min_lr + coeff * delta_lr

    def step_absolute(self, num_steps):
        """Set lr for all parameters groups."""
        if num_steps is None:
            self.num_steps += 1
        else:
            self.num_steps = num_steps
        new_wd = self.get_wd()
        for param_group in self.optimizer.param_groups:
            new_lr = self.get_lr(param_group)
            param_group["lr"] = new_lr * param_group.get("lr_mult", 1.0)
            param_group["weight_decay"] = new_wd * param_group.get("wd_mult", 1.0)

    def step(self, increment):
        """Set lr for all parameters groups."""
        self.num_steps += increment
        new_wd = self.get_wd()
        for param_group in self.optimizer.param_groups:
            new_lr = self.get_lr(param_group)
            param_group["lr"] = new_lr * param_group.get("lr_mult", 1.0)
            param_group["weight_decay"] = new_wd * param_group.get("wd_mult", 1.0)

    def state_dict(self):
        state_dict = {
            "max_lr": self.max_lr,
            "lr_warmup_steps": self.lr_warmup_steps,
            "num_steps": self.num_steps,
            "lr_decay_style": self.lr_decay_style,
            "lr_decay_steps": self.lr_decay_steps,
            "min_lr": self.min_lr,
            "start_wd": self.start_wd,
            "end_wd": self.end_wd,
            "wd_incr_style": self.wd_incr_style,
            "wd_incr_steps": self.wd_incr_steps,
        }
        return state_dict

    def _check_and_set(self, cls_value, sd_value, name):
        """Auxiliary function for checking the values in the checkpoint and
        setting them."""
        if self.override_opt_param_scheduler:
            self.log_rank_0(" > overriding {} value to {}".format(name, cls_value))
            return cls_value

        if not self.use_checkpoint_opt_param_scheduler:
            assert cls_value == sd_value, (
                f"OptimizerParamScheduler: class input value {cls_value} and checkpoint"
                f"value {sd_value} for {name} do not match"
            )
        self.log_rank_0(" > using checkpoint value {} for {}".format(sd_value, name))
        return sd_value

    def load_state_dict(self, sd):

        if "start_lr" in sd:
            max_lr_ = sd["start_lr"]
        else:
            max_lr_ = sd["max_lr"]
        self.max_lr = self._check_and_set(self.max_lr, max_lr_, "learning rate")

        self.min_lr = self._check_and_set(
            self.min_lr, sd["min_lr"], "minimum learning rate"
        )

        if "warmup_iter" in sd:
            lr_warmup_steps_ = sd["warmup_iter"]
        elif "warmup_steps" in sd:
            lr_warmup_steps_ = sd["warmup_steps"]
        else:
            lr_warmup_steps_ = sd["lr_warmup_steps"]
        self.lr_warmup_steps = self._check_and_set(
            self.lr_warmup_steps, lr_warmup_steps_, "warmup iterations"
        )

        if "end_iter" in sd:
            lr_decay_steps_ = sd["end_iter"]
        elif "decay_steps" in sd:
            lr_decay_steps_ = sd["decay_steps"]
        else:
            lr_decay_steps_ = sd["lr_decay_steps"]
        self.lr_decay_steps = self._check_and_set(
            self.lr_decay_steps, lr_decay_steps_, "total number of iterations"
        )

        if "decay_style" in sd:
            lr_decay_style_ = sd["decay_style"]
        else:
            lr_decay_style_ = sd["lr_decay_style"]
        self.lr_decay_style = self._check_and_set(
            self.lr_decay_style, lr_decay_style_, "learning rate decay style"
        )

        if "num_iters" in sd:
            num_steps = sd["num_iters"]
        else:
            num_steps = sd["num_steps"]
        self.step(increment=num_steps)

        if "start_wd" in sd:
            self.start_wd = self._check_and_set(
                self.start_wd, sd["start_wd"], "start weight decay"
            )
            self.end_wd = self._check_and_set(
                self.end_wd, sd["end_wd"], "end weight decay"
            )
            self.wd_incr_steps = self._check_and_set(
                self.wd_incr_steps,
                sd["wd_incr_steps"],
                "total number of weight decay iterations",
            )
            self.wd_incr_style = self._check_and_set(
                self.wd_incr_style,
                sd["wd_incr_style"],
                "weight decay incr style",
            )


@torch.no_grad()
def _step_megatron_distrib_optimizer_internal(optim: MegatronDistOptim):
    # NOTE: patching this function to use the correct model parallel group

    optim._copy_model_grads_to_main_grads()

    # Do unscale, check for inf, and update grad scaler only for
    # the case that grad scaler is provided.
    if optim.grad_scaler:

        def _unscale_main_grads_and_check_for_nan(optim: MegatronDistOptim):

            # Collect main grads.
            main_grads = optim._collect_main_grad_data_for_unscaling()

            # Reset found inf.
            optim.found_inf.fill_(0.0)

            # Unscale and set found inf/nan
            torch._amp_foreach_non_finite_check_and_unscale_(
                main_grads, optim.found_inf, optim.grad_scaler.inv_scale
            )

            # Update across all model parallel instances.
            dist.all_reduce(
                optim.found_inf,
                op=dist.ReduceOp.MAX,
                group=constants.parallelism_group(),
            )

            # Check for nan.
            found_inf_flag = optim.found_inf.item() > 0

            return found_inf_flag

        # Unscale and check for inf/nan.
        found_inf_flag = _unscale_main_grads_and_check_for_nan(optim)

        # We are done with scaling gradients
        # so we can update the loss scale.
        optim.grad_scaler.update(found_inf_flag)

        # If we found inf/nan, skip the update.
        if found_inf_flag:
            optim.update_successful, grad_norm, num_zeros_in_grad = (
                False,
                None,
                None,
            )
            return optim.update_successful, grad_norm, num_zeros_in_grad

    def clip_grad_norm(optim: MegatronDistOptim, clip_grad: float) -> float:
        """Compute grad norm."""
        params = optim.get_parameters()
        grads_for_norm = optim.get_main_grads_for_grad_norm()
        return clip_grad_norm_fp32(
            params,
            grads_for_norm,
            clip_grad,
            model_parallel_group=constants.parallelism_group(),
        )

    def count_zeros(optim: MegatronDistOptim) -> float:
        """Count number of zeros in model's gradients."""
        params = optim.get_parameters()
        return count_zeros_fp32(
            params,
            model_parallel_group=constants.parallelism_group(),
        )

    # Clip the main gradients.
    grad_norm = None
    if optim.config.clip_grad > 0.0:
        grad_norm = clip_grad_norm(optim, optim.config.clip_grad)

    # Count the zeros in the grads.
    num_zeros_in_grad = None
    if optim.config.log_num_zeros_in_grad:
        num_zeros_in_grad = count_zeros(optim)

    # Step the optimizer.
    optim.optimizer.step()

    # Update params from main params.
    optim._copy_main_params_to_model_params()
    # Successful update.
    optim.update_successful, grad_norm, num_zeros_in_grad = (
        True,
        grad_norm,
        num_zeros_in_grad,
    )
    return optim.update_successful, grad_norm, num_zeros_in_grad


def step_megatron_distrb_optimizer(optim: MegatronDistOptim):

    optim.update_successful, grad_norm, num_zeros_in_grad = (
        _step_megatron_distrib_optimizer_internal(optim)
    )

    # If not overlapping all-gather for parameters, launch synchronous all-gather
    # communication calls here. If overlapping all-gather for parameters, the following
    # call to _gather_all_model_params is a no-op: the first all-gather is launched
    # asynchronously in the next optimizer.zero_grad() call and subsequent all-gathers
    # are launched in the forward pre-hook.
    optim._reset_metadata_and_sync_gather_all_model_params(force_sync=False)

    return optim.update_successful, grad_norm, num_zeros_in_grad


def _flatten_dense_tensors(tensors):
    """Flatten dense tensors into a contiguous 1D buffer. Assume tensors are of
    same dense type.

    Since inputs are dense, the resulting tensor will be a concatenated 1D
    buffer. Element-wise operation on this buffer will be equivalent to
    operating individually.

    Args:
        tensors (Iterable[Tensor]): dense tensors to flatten.

    Returns:
        A contiguous 1D buffer containing input tensors.
    """
    return torch._C._nn.flatten_dense_tensors(tensors)


def _unflatten_dense_tensors(flat, tensors):
    """View a flat buffer using the sizes of tensors. Assume that tensors are of
    same dense type, and that flat is given by _flatten_dense_tensors.

    Args:
        flat (Tensor): flattened dense tensors to unflatten.
        tensors (Iterable[Tensor]): dense tensors whose sizes will be used to
          unflatten flat.

    Returns:
        Unflattened dense tensors with sizes same as tensors and values from
        flat.
    """
    return torch._C._nn.unflatten_dense_tensors(flat, tensors)


def megatron_all_reduce_layernorm_grads(engine: MegatronEngine):
    if not (
        constants.sequence_parallel() and constants.model_parallel_world_size() > 1
    ):
        return
    real_model: ReaLModel = engine.ddp.module
    grads = []
    for i in range(real_model.layer_idx_start, real_model.layer_idx_end):
        if i == 0:
            continue
        elif i == real_model.config.n_layers + 1:
            continue
        else:
            assert 0 < i < real_model.config.n_layers + 1
            layer: ReaLModelBlock = real_model.layers[i - real_model.layer_idx_start]
            grads.append(layer.attn.c_attn.ln.weight.main_grad)
            if getattr(layer.attn.c_attn.ln, "bias", None) is not None:
                grads.append(layer.attn.c_attn.ln.bias.main_grad)
            grads.append(layer.mlp.ln.weight.main_grad)
            if getattr(layer.mlp.ln, "bias", None) is not None:
                grads.append(layer.mlp.ln.bias.main_grad)
            if i == real_model.config.n_layers:
                grads.append(layer.ln_f.weight.main_grad)
                if getattr(layer.ln_f, "bias", None) is not None:
                    grads.append(layer.ln_f.bias.main_grad)

    assert all(x is not None for x in grads)
    coalesced = _flatten_dense_tensors(grads)
    dist.all_reduce(coalesced, group=constants.model_parallel_group())
    for buf, synced in zip(grads, _unflatten_dense_tensors(coalesced, grads)):
        buf.copy_(synced)


def megatron_all_reduce_word_embedding_grads(engine: MegatronEngine):
    real_model: ReaLModel = engine.ddp.module
    if not real_model.config.tied_embedding or real_model.config.is_critic:
        return
    pp_size = constants.pipe_parallel_world_size()
    pp_rank = constants.pipe_parallel_rank()
    if pp_size == 1:
        return
    if pp_rank not in [0, pp_size - 1]:
        return

    if pp_rank == 0:
        grad = real_model.layers[0].wte.weight.main_grad
    else:
        grad = real_model.layers[-1].weight.main_grad

    dist.all_reduce(grad, group=constants.grid().embedding_proc_group)


def finalize_grads_megatron(engine: MegatronEngine):
    engine.ddp.finish_grad_sync()
    megatron_all_reduce_layernorm_grads(engine)
    megatron_all_reduce_word_embedding_grads(engine)



import abc
import copy
import dataclasses
import keyword
import os
from collections import defaultdict
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Union

import torch
import torch.utils.data
import transformers

import realhf.base.logging as logging
from realhf.api.core import dfg, system_api
from realhf.api.core.config import ModelFamily, ModelName
from realhf.base.namedarray import NamedArray

logger = logging.getLogger("model")


@dataclasses.dataclass
class GenerationHyperparameters:
    """Generation hyperparameters.

    We implement a customized generation function instead of
    using HuggingFace's to support pipelined generation.
    As a result, advanced generation techniques like
    diversity-promoting sampling or repeatition penalty
    are not supported during PPO training.
    However, we don't find it to be a problem in practice.
    Increasing the sampling temperature and enabling
    top-k/top-p sampling can produce good models.

    :param max_new_tokens: The maximum number of new tokens to generate.
    :type max_new_tokens: int
    :param min_new_tokens: The minimum number of new tokens to generate.
    :type min_new_tokens: int
    :param greedy: Whether to use greedy decoding.
    :type greedy: bool
    :param top_k: The number of highest probability tokens to keep.
    :type top_k: int
    :param top_p: The cumulative probability of the highest probability tokens to keep.
    :type top_p: float
    :param temperature: The temperature of the sampling process.
    :type temperature: float
    :param num_samples: The number of samples to generate.
        Should only be enabled for pure generation.
        Must be 1 for PPO generation.
    :type num_samples: int
    :param use_cuda_graph: Whether to use CUDA graph to reduce kernel launch overhead
        during generation. Recommended for pure generation.
    :type use_cuda_graph: bool
    """

    max_new_tokens: int = 256
    min_new_tokens: int = 256
    greedy: bool = False
    top_p: float = 0.9
    top_k: int = 200
    temperature: float = 1.0
    num_samples: int = 1
    use_cuda_graph: bool = False


@dataclasses.dataclass
class ReaLModelConfig:
    """Configuration for ReaLModel.

    :param n_layers: Number of transformer blocks.
    :type n_layers: int
    :param n_kv_heads: Number of key-value attention heads.
    :type n_kv_heads: int
    :param n_q_heads: Number of query attention heads.
    :type n_q_heads: int
    :param head_dim: Dimension of each attention head.
        If None, defaults to hidden_dim // n_q_heads.
        If given, the query layer will have shape (hidden_dim, head_dim * n_q_heads).
    :type head_dim: int
    :param hidden_dim: Hidden dimension of the transformer block.
    :type hidden_dim: int
    :param intermediate_dim: Dimension of the intermediate layer in the MLP.
    :type intermediate_dim: int
    :param vocab_size: Vocabulary size.
    :type vocab_size: int
    :param n_positions: Maximum context length. Can be None for
        rotary embedding, where the context length is decided during runtime.
    :type n_positions: Optional[int]
    :param embd_pdrop: Dropout probability for the embedding layer.
    :type embd_pdrop: float
    :param resid_pdrop: Dropout probability for the residual connections.
    :type resid_pdrop: float
    :param attn_pdrop: Dropout probability for the attention weights.
    :type attn_pdrop: float
    :param layer_norm_epsilon: Epsilon for layer normalization.
    :type layer_norm_epsilon: float
    :param activation_function: Activation function for the MLP.
    :type activation_function: str
    :param scale_attn_by_inverse_layer_idx: Whether to scale the attention weights
        by the inverse of the layer index.
    :type scale_attn_by_inverse_layer_idx: bool
    :param use_attention_bias: Whether to use bias for QKV layers.
    :type use_attention_bias: bool
    :param use_attn_proj_bias: Whether to use bias for the attention projection layer.
    :type use_attn_proj_bias: bool
    :param layer_norm_type: Type of layer normalization, can by None, "rms", or "gemma".
    :type layer_norm_type: Optional[str]
    :param mlp_type: Type of the MLP. Either None or "llama".
    :type mlp_type: Optional[str]
    :param apply_rotary: Whether to apply rotary embedding.
    :type apply_rotary: bool
    :param rotary_base: Exponential base for the rotary embedding.
    :type rotary_base: float
    :param rotary_interleaved: Whether to use interleaved rotary embedding.
    :type rotary_interleaved: bool
    :param rotary_scaling: Scaling factor for the rotary embedding.
    :type rotary_scaling: Optional[float]
    :param rotary_scaling_type: Type of scaling for the rotary embedding.
    :type rotary_scaling_type: Optional[str]
    :param normalize_embed: Whether to normalize the embeddings
        before transformer blocks. Used by Gemma.
    :type normalize_embed: bool
    :param abs_position_embedding_offset: Offset for the absolute position embedding.
        Used by OPT, but OPT is currently not supported.
    :type abs_position_embedding_offset: int
    :param do_layernorm_before: Whether to apply layer normalization before the attention
        rather than after. Used by OPT, but OPT is currently not supported.
    :type do_layernorm_before: bool
    :param tied_embedding: Whether to share the embeddings and output weights.
        Used by models like GPT-2 and Gemma.
    :type tied_embedding: bool
    :param sliding_window: Sliding window size for the attention.
        Currently a placeholder and not supported.
    :type sliding_window: Optional[int]
    :param is_critic: Whether the model is a critic model.
    :type is_critic: bool
    :param gradient_accumulation_fusion: Whether to fuse
        gradient accumulation in Megatron.
        Currently not supported.
    :type gradient_accumulation_fusion: bool
    """

    ### Architectural configurations. ###
    n_layers: int
    n_kv_heads: int
    n_q_heads: int
    hidden_dim: int
    intermediate_dim: int  # for mlp, usually 4*h
    vocab_size: int
    head_dim: Optional[int] = None
    n_positions: Optional[int] = None
    embd_pdrop: float = 0.1
    resid_pdrop: float = 0.1
    attn_pdrop: float = 0.1
    layer_norm_epsilon: float = 1e-5
    activation_function: str = "gelu"
    scale_attn_by_inverse_layer_idx: bool = True
    scale_attn_weights: bool = True
    # llama does not use attention bias and uses special MLP/LayerNorm layers
    use_attention_bias: bool = True
    use_attn_proj_bias: bool = True
    layer_norm_type: Optional[str] = None
    mlp_type: Optional[str] = None
    # rotary embedding
    apply_rotary: bool = False
    rotary_base: float = 10000.0
    rotary_interleaved: bool = False
    rotary_scaling: Optional[float] = None
    rotary_scaling_type: Optional[str] = None
    # for gemma
    normalize_embed: bool = False
    # for opt, it's 2
    abs_position_embedding_offset: int = 0
    do_layernorm_before: bool = True
    # Tied embedding
    tied_embedding: bool = False
    sliding_window: Optional[int] = None
    # Whether it is a critic/reward model that outputs scores.
    is_critic: bool = False

    ### Running configurations. ###
    gradient_accumulation_fusion: bool = False

    def __post_init__(self):
        if self.is_critic and self.tied_embedding:
            raise ValueError("Critic model cannot share embeddings and output weights.")
        if self.head_dim is None:
            self.head_dim = self.hidden_dim // self.n_q_heads


def load_hf_tokenizer(
    model_name_or_path: str,
    fast_tokenizer=True,
    padding_side: Optional[str] = None,
) -> transformers.PreTrainedTokenizerFast:
    kwargs = {}
    if padding_side is not None:
        kwargs["padding_side"] = padding_side
    tokenizer = transformers.AutoTokenizer.from_pretrained(
        model_name_or_path, fast_tokenizer=fast_tokenizer, **kwargs
    )
    if tokenizer.pad_token_id is None:
        tokenizer.pad_token_id = tokenizer.eos_token_id
    return tokenizer


@dataclasses.dataclass
class ModelVersion:
    epoch: int = 0
    epoch_step: int = 0
    global_step: int = 0


@dataclasses.dataclass
class FinetuneSpec:
    total_train_epochs: int
    total_train_steps: int
    steps_per_epoch: int


@dataclasses.dataclass
class Model:
    name: ModelName
    module: torch.nn.Module
    tokenizer: transformers.PreTrainedTokenizerFast
    device: Union[str, torch.device]
    dtype: Optional[torch.dtype] = None
    version: ModelVersion = dataclasses.field(default_factory=ModelVersion)
    ft_spec: FinetuneSpec = None  # will be initialized by the backend

    def __post_init__(self):
        try:
            self.module = self.module.to(self.device)
        except ValueError as e:
            # 4-bit and 8-bit model may fail here
            logger.warning(
                f"Failed to move model to device {self.device} because {e}. Abort to device."
            )

    def inc_version(self):
        self.version.global_step += 1
        self.version.epoch_step += 1
        if self.version.epoch_step >= self.ft_spec.steps_per_epoch:
            self.version.epoch_step = 0
            self.version.epoch += 1


class ModelBackend(abc.ABC):

    @abc.abstractmethod
    def _initialize(self, model: Model, spec: FinetuneSpec) -> Model:
        raise NotImplementedError()

    def initialize(self, model: Model, spec: FinetuneSpec) -> Model:
        model.ft_spec = spec
        return self._initialize(model, spec)


class NullBackend(ModelBackend):

    def _initialize(self, model: Model, spec: FinetuneSpec) -> Model:
        return model


def null_model(name: ModelName, device: Union[str, torch.device]) -> Model:
    return Model(name, torch.nn.Identity(), None, device)


def tokenizer_only_model(
    name: ModelName, device: Union[str, torch.device], tokenizer_path: str
) -> Model:
    return Model(name, torch.nn.Identity(), load_hf_tokenizer(tokenizer_path), device)


class ModelInterface(abc.ABC):
    """Interface for model training, evaluation, inference and generation."""

    def save(self, model: Model, save_dir: str):
        pass

    def evaluate(
        self, model: Model, eval_dataloader: torch.utils.data.DataLoader
    ) -> Dict:
        return {}

    def inference(self, model: Model, data: NamedArray) -> NamedArray:
        raise NotImplementedError()

    def generate(self, model: Model, data: NamedArray) -> NamedArray:
        raise NotImplementedError()

    def train_step(self, model: Model, data: NamedArray) -> Dict:
        raise NotImplementedError()


ALL_MODEL_CLASSES = {}
ALL_INTERFACE_CLASSES = {}
ALL_BACKEND_CLASSES = {}
ALL_WRAPPER_CLASSES = {}


def register_model(name, model_cls):
    assert name not in ALL_MODEL_CLASSES
    ALL_MODEL_CLASSES[name] = model_cls


def register_interface(name, cls_):
    assert name not in ALL_INTERFACE_CLASSES
    assert issubclass(cls_, ModelInterface)
    ALL_INTERFACE_CLASSES[name] = cls_


def register_backend(name, cls_):
    assert name not in ALL_BACKEND_CLASSES
    assert issubclass(cls_, ModelBackend)
    ALL_BACKEND_CLASSES[name] = cls_


def register_wrapper(name, cls_):
    assert name not in ALL_WRAPPER_CLASSES
    ALL_WRAPPER_CLASSES[name] = cls_


def make_model_wrapper(
    cfg: system_api.ModelWrapper,
) -> Callable[[Model], Model]:
    cls_ = ALL_WRAPPER_CLASSES[cfg.type_]
    return cls_(**cfg.args)


def make_model(
    cfg: system_api.Model, name: ModelName, device: Union[str, torch.device]
) -> Model:
    logger.debug(f"making model {cfg.type_} on {device}")
    model_cls = ALL_MODEL_CLASSES[cfg.type_]
    model = model_cls(**cfg.args, name=name, device=device)
    assert isinstance(model, Model)
    for w in cfg.wrappers:
        model = make_model_wrapper(w)(model)
        assert isinstance(model, Model)
    return model


def make_interface(cfg: dfg.ModelInterface) -> ModelInterface:
    cls_ = ALL_INTERFACE_CLASSES[cfg.type_]
    return cls_(**cfg.args)


def make_backend(cfg: system_api.ModelBackend) -> ModelBackend:
    cls_ = ALL_BACKEND_CLASSES[cfg.type_]
    return cls_(**cfg.args)


register_backend("null", NullBackend)
register_model("null", null_model)
register_model("tokenizer", tokenizer_only_model)

SUPPORTED_MODELS = []
HF_MODEL_FAMILY_REGISTRY = {}


def is_valid_function_name(name):
    if not name.isidentifier():
        return False
    if keyword.iskeyword(name):
        return False
    return True


def register_hf_family(
    name: str,
    hf_cls_name: str,
    config_from_hf_converter: Callable[
        [transformers.PretrainedConfig], ReaLModelConfig
    ],
    config_to_hf_converter: Callable[[ReaLModelConfig], transformers.PretrainedConfig],
    sd_from_hf_converter: Callable[[Dict, ReaLModelConfig], Dict],
    sd_to_hf_converter: Callable[[Dict, ReaLModelConfig], Dict],
    embedding_param_names: Callable[[ReaLModelConfig], List[str]],
    tblock_param_names: Callable[[ReaLModelConfig, int], List[str]],
    head_param_names: Callable[[ReaLModelConfig], List[str]],
    real_config_maker: Optional[Callable] = None,
):
    if name in SUPPORTED_MODELS:
        raise ValueError(f"Model {name} is already registered.")
    if not is_valid_function_name(name):
        raise ValueError(f"Model name {name} is not a valid function name in Python.")
    SUPPORTED_MODELS.append(name)
    HF_MODEL_FAMILY_REGISTRY[name] = dict(
        name=name,
        hf_cls_name=hf_cls_name,
        config_from_hf_converter=config_from_hf_converter,
        config_to_hf_converter=config_to_hf_converter,
        sd_from_hf_converter=sd_from_hf_converter,
        sd_to_hf_converter=sd_to_hf_converter,
        embedding_param_names=embedding_param_names,
        tblock_param_names=tblock_param_names,
        head_param_names=head_param_names,
        real_config_maker=real_config_maker,
    )



import time

import pytest
import torch

from realhf.impl.model.utils.ppo_functional import (
    cugae1d_nolp_misalign_func,
    cugae2d_nolp_func,
    cugae2d_olp_func,
    pygae1d_nolp_misalign,
    pygae2d_nolp,
    pygae2d_olp,
)


@pytest.mark.skipif(not torch.cuda.is_available(), reason="This test requires a GPU.")
@pytest.mark.parametrize("max_seqlen", [32, 128, 512])
@pytest.mark.parametrize("bs", [2, 4])
@pytest.mark.parametrize("gamma", [0.9, 1.0])
@pytest.mark.parametrize("lam", [0.5, 1.0])
def test_gae1d_nolp_misalign(max_seqlen: int, bs: int, gamma: float, lam: float):

    seqlens = torch.randint(1, max_seqlen, (bs,), dtype=torch.int32, device="cuda")
    rewards = torch.randn(seqlens.sum(), dtype=torch.float32, device="cuda")
    values = torch.randn(seqlens.sum() + bs, dtype=torch.float32, device="cuda")
    bootstrap = torch.ones(bs, dtype=torch.bool, device="cuda")
    cu_seqlens = torch.nn.functional.pad(seqlens.cumsum(0), (1, 0)).int()

    adv, ret = cugae1d_nolp_misalign_func(
        rewards, values, cu_seqlens, bootstrap, gamma, lam
    )
    py_adv, py_ret = pygae1d_nolp_misalign(
        rewards, values, cu_seqlens, bootstrap, gamma, lam
    )

    torch.cuda.synchronize()
    t1 = time.perf_counter_ns()
    py_adv, py_ret = pygae1d_nolp_misalign(
        rewards, values, cu_seqlens, bootstrap, gamma, lam
    )
    torch.cuda.synchronize()
    t2 = time.perf_counter_ns()
    adv, ret = cugae1d_nolp_misalign_func(
        rewards, values, cu_seqlens, bootstrap, gamma, lam
    )
    torch.cuda.synchronize()
    t3 = time.perf_counter_ns()

    assert torch.allclose(adv, py_adv, atol=1e-5), (adv - py_adv).abs().max()
    assert torch.allclose(ret, py_ret, atol=1e-5), (ret - py_ret).abs().max()

    print(
        f"max_seqlen={max_seqlen},bs={bs}, CUDA acceleration ratio",
        (t2 - t1) / (t3 - t2),
    )


@pytest.mark.skipif(not torch.cuda.is_available(), reason="This test requires a GPU.")
@pytest.mark.parametrize("seqlen", [32, 128, 512, 1024])
@pytest.mark.parametrize("bs", [8, 16, 32, 100])
@pytest.mark.parametrize("gamma", [0.9, 1.0])
@pytest.mark.parametrize("lam", [0.5, 1.0])
def test_gae2d_olp(bs: int, seqlen: int, gamma: float, lam: float):

    torch.random.manual_seed(0)
    rewards = torch.randn(bs, seqlen).cuda()
    values = torch.randn(bs, seqlen + 1).cuda()
    dones = torch.randint(0, 2, (bs, seqlen + 1)).bool().cuda()
    truncates = dones.logical_and(torch.randint(0, 2, (bs, seqlen + 1)).bool().cuda())

    py_adv, py_ret = pygae2d_olp(rewards, values, dones, truncates, gamma, lam)
    adv, ret = cugae2d_olp_func(rewards, values, dones, truncates, gamma, lam)

    torch.cuda.synchronize()
    t1 = time.perf_counter_ns()
    py_adv, py_ret = pygae2d_olp(rewards, values, dones, truncates, gamma, lam)
    torch.cuda.synchronize()
    t2 = time.perf_counter_ns()
    adv, ret = cugae2d_olp_func(rewards, values, dones, truncates, gamma, lam)
    torch.cuda.synchronize()
    t3 = time.perf_counter_ns()

    assert torch.allclose(adv, py_adv, atol=1e-5), (adv - py_adv).abs().max()
    assert torch.allclose(ret, py_ret, atol=1e-5), (ret - py_ret).abs().max()
    print(
        f"seqlen={seqlen},bs={bs}, CUDA acceleration ratio",
        (t2 - t1) / (t3 - t2),
    )


@pytest.mark.skipif(not torch.cuda.is_available(), reason="This test requires a GPU.")
@pytest.mark.parametrize("seqlen", [32, 128, 512, 1024])
@pytest.mark.parametrize("bs", [8, 16, 32, 100])
@pytest.mark.parametrize("gamma", [0.9, 1.0])
@pytest.mark.parametrize("lam", [0.5, 1.0])
def test_gae2d_nolp(bs: int, seqlen: int, gamma: float, lam: float):

    torch.random.manual_seed(0)
    rewards = torch.randn(bs, seqlen).cuda()
    values = torch.randn(bs, seqlen + 1).cuda()
    on_reset_ = torch.randint(0, 2, (bs, seqlen + 2)).bool().cuda()
    on_reset = on_reset_[:, :-1].contiguous()
    truncates = (
        on_reset_[:, 1:]
        .logical_and(torch.randint(0, 2, (bs, seqlen + 1)).bool().cuda())
        .contiguous()
    )

    py_adv, py_ret = pygae2d_nolp(rewards, values, on_reset, truncates, gamma, lam)
    adv, ret = cugae2d_nolp_func(rewards, values, on_reset, truncates, gamma, lam)

    torch.cuda.synchronize()
    t1 = time.perf_counter_ns()
    py_adv, py_ret = pygae2d_nolp(rewards, values, on_reset, truncates, gamma, lam)
    torch.cuda.synchronize()
    t2 = time.perf_counter_ns()
    adv, ret = cugae2d_nolp_func(rewards, values, on_reset, truncates, gamma, lam)
    torch.cuda.synchronize()
    t3 = time.perf_counter_ns()

    adv = adv * (1 - truncates[:, :-1].float())
    py_adv = py_adv * (1 - truncates[:, :-1].float())
    ret = ret * (1 - truncates[:, :-1].float())
    py_ret = py_ret * (1 - truncates[:, :-1].float())
    assert torch.allclose(adv, py_adv, atol=1e-5), (adv - py_adv).abs().max()
    assert torch.allclose(ret, py_ret, atol=1e-5), (ret - py_ret).abs().max()
    print(
        f"seqlen={seqlen},bs={bs}, CUDA acceleration ratio",
        (t2 - t1) / (t3 - t2),
    )



import dataclasses
from typing import Dict, Optional, Tuple

import torch

import realhf.api.core.model_api as model_api
from realhf.base.namedarray import NamedArray, recursive_apply


@dataclasses.dataclass
class GenerationInterface(model_api.ModelInterface):
    generation_config: model_api.GenerationHyperparameters = dataclasses.field(
        default_factory=model_api.GenerationHyperparameters
    )

    @torch.no_grad()
    def generate(self, model: model_api.Model, data: NamedArray) -> NamedArray:
        module = model.module

        module.eval()

        data = recursive_apply(data, lambda x: x.to(model.device))
        packed_prompts = data["packed_prompts"]
        prompt_lengths = torch.tensor(
            data.metadata["seqlens"],
            dtype=torch.int32,
            device=packed_prompts.device,
        )
        prompt_cu_seqlens = torch.nn.functional.pad(prompt_lengths.cumsum(0), (1, 0))

        res = module.generate(
            seqlens_cpu=data.metadata["seqlens"],
            tokenizer=model.tokenizer,
            packed_input_ids=packed_prompts,
            cu_seqlens=prompt_cu_seqlens,
            gconfig=self.generation_config,
        )
        if res is None:
            return None

        gen_tokens, logprobs, *_ = res
        res = {
            "generated_length": gen_tokens.shape[1],
            "batch_size": gen_tokens.shape[0],
        }
        return res


model_api.register_interface("generation", GenerationInterface)



import os
from typing import *

import torch
import transformers

from realhf.api.core.model_api import ReaLModelConfig, register_hf_family


def sd_from_gpt2(state_dict: Dict, config: ReaLModelConfig) -> Dict:
    if any(k.startswith("transformer.") for k in state_dict.keys()):
        new_sd = {k.replace("transformer.", ""): v for k, v in state_dict.items()}
        if "lm_head.weight" in new_sd:
            head_w = new_sd.pop("lm_head.weight")
            assert torch.allclose(head_w, new_sd["wte.weight"])
        state_dict = new_sd

    new_sd = {}
    if "wte.weight" in state_dict:
        new_sd["0.wte.weight"] = state_dict["wte.weight"]
    if "wpe.weight" in state_dict:
        new_sd["0.wpe.weight"] = state_dict["wpe.weight"]
    if "lm_head.weight" in state_dict:
        new_sd[f"{config.n_layers + 1}.weight"] = state_dict["lm_head.weight"]
    if "ln_f.weight" in state_dict:
        new_sd[f"{config.n_layers}.ln_f.weight"] = state_dict["ln_f.weight"]
        new_sd[f"{config.n_layers}.ln_f.bias"] = state_dict["ln_f.bias"]
    for i in range(config.n_layers):
        if not any(k.startswith(f"h.{i}.") for k in state_dict):
            continue
        new_sd[f"{i+1}.attn.c_attn.ln.weight"] = state_dict[f"h.{i}.ln_1.weight"]
        new_sd[f"{i+1}.attn.c_attn.ln.bias"] = state_dict[f"h.{i}.ln_1.bias"]

        attn_w = state_dict[f"h.{i}.attn.c_attn.weight"]
        attn_bias = state_dict[f"h.{i}.attn.c_attn.bias"]
        qw, kw, vw = torch.chunk(attn_w, 3, dim=1)
        qb, kb, vb = torch.chunk(attn_bias, 3, dim=0)
        new_sd[f"{i+1}.attn.c_attn.q_attn.weight"] = qw.transpose(0, 1)
        new_sd[f"{i+1}.attn.c_attn.k_attn.weight"] = kw.transpose(0, 1)
        new_sd[f"{i+1}.attn.c_attn.v_attn.weight"] = vw.transpose(0, 1)
        new_sd[f"{i+1}.attn.c_attn.q_attn.bias"] = qb
        new_sd[f"{i+1}.attn.c_attn.k_attn.bias"] = kb
        new_sd[f"{i+1}.attn.c_attn.v_attn.bias"] = vb

        new_sd[f"{i+1}.attn.c_proj.weight"] = state_dict[
            f"h.{i}.attn.c_proj.weight"
        ].transpose(0, 1)
        new_sd[f"{i+1}.attn.c_proj.bias"] = state_dict[f"h.{i}.attn.c_proj.bias"]

        new_sd[f"{i+1}.mlp.ln.weight"] = state_dict[f"h.{i}.ln_2.weight"]
        new_sd[f"{i+1}.mlp.ln.bias"] = state_dict[f"h.{i}.ln_2.bias"]
        new_sd[f"{i+1}.mlp.c_fc.weight"] = state_dict[
            f"h.{i}.mlp.c_fc.weight"
        ].transpose(0, 1)
        new_sd[f"{i+1}.mlp.c_fc.bias"] = state_dict[f"h.{i}.mlp.c_fc.bias"]
        new_sd[f"{i+1}.mlp.c_proj.weight"] = state_dict[
            f"h.{i}.mlp.c_proj.weight"
        ].transpose(0, 1)
        new_sd[f"{i+1}.mlp.c_proj.bias"] = state_dict[f"h.{i}.mlp.c_proj.bias"]
    return new_sd


def sd_to_gpt2(state_dict: Dict, config: ReaLModelConfig) -> Dict:
    max_positions = config.n_positions
    bias = torch.tril(
        torch.ones((max_positions, max_positions), dtype=torch.bool)
    ).view(1, 1, max_positions, max_positions)

    new_sd = {}
    if "0.wte.weight" in state_dict:
        new_sd["wte.weight"] = state_dict["0.wte.weight"]
        new_sd["wpe.weight"] = state_dict["0.wpe.weight"]
    if f"{config.n_layers}.ln_f.weight" in state_dict:
        new_sd["ln_f.weight"] = state_dict[f"{config.n_layers}.ln_f.weight"]
        new_sd["ln_f.bias"] = state_dict[f"{config.n_layers}.ln_f.bias"]
    if f"{config.n_layers + 1}.weight" in state_dict:
        new_sd["lm_head.weight"] = state_dict[f"{config.n_layers + 1}.weight"]
    for i in range(config.n_layers):
        if not any(k.startswith(f"{i+1}.") for k in state_dict):
            continue
        new_sd[f"h.{i}.ln_1.weight"] = state_dict[f"{i+1}.attn.c_attn.ln.weight"]
        new_sd[f"h.{i}.ln_1.bias"] = state_dict[f"{i+1}.attn.c_attn.ln.bias"]

        qw = state_dict[f"{i+1}.attn.c_attn.q_attn.weight"].transpose(0, 1)
        kw = state_dict[f"{i+1}.attn.c_attn.k_attn.weight"].transpose(0, 1)
        vw = state_dict[f"{i+1}.attn.c_attn.v_attn.weight"].transpose(0, 1)
        qb = state_dict[f"{i+1}.attn.c_attn.q_attn.bias"]
        kb = state_dict[f"{i+1}.attn.c_attn.k_attn.bias"]
        vb = state_dict[f"{i+1}.attn.c_attn.v_attn.bias"]
        attn_w = torch.cat([qw, kw, vw], dim=1)
        attn_bias = torch.cat([qb, kb, vb], dim=0)
        new_sd[f"h.{i}.attn.c_attn.weight"] = attn_w
        new_sd[f"h.{i}.attn.c_attn.bias"] = attn_bias

        new_sd[f"h.{i}.attn.c_proj.weight"] = state_dict[
            f"{i+1}.attn.c_proj.weight"
        ].transpose(0, 1)
        new_sd[f"h.{i}.attn.c_proj.bias"] = state_dict[f"{i+1}.attn.c_proj.bias"]

        new_sd[f"h.{i}.ln_2.weight"] = state_dict[f"{i+1}.mlp.ln.weight"]
        new_sd[f"h.{i}.ln_2.bias"] = state_dict[f"{i+1}.mlp.ln.bias"]
        new_sd[f"h.{i}.mlp.c_fc.weight"] = state_dict[
            f"{i+1}.mlp.c_fc.weight"
        ].transpose(0, 1)
        new_sd[f"h.{i}.mlp.c_fc.bias"] = state_dict[f"{i+1}.mlp.c_fc.bias"]
        new_sd[f"h.{i}.mlp.c_proj.weight"] = state_dict[
            f"{i+1}.mlp.c_proj.weight"
        ].transpose(0, 1)
        new_sd[f"h.{i}.mlp.c_proj.bias"] = state_dict[f"{i+1}.mlp.c_proj.bias"]
        new_sd[f"h.{i}.attn.bias"] = bias
    return new_sd


# param name is used to load directly from huggingface checkpoints
def gpt2_embedding_layer_names(config: ReaLModelConfig) -> List[str]:
    return [
        "wte.weight",
        "wpe.weight",
        "transformer.wte.weight",
        "transformer.wpe.weight",
    ]


def gpt2_transformer_block_param_name(config: ReaLModelConfig, idx: int) -> List[str]:
    names = [
        f"h.{idx}.ln_1.weight",
        f"h.{idx}.ln_1.bias",
        # f'h.{idx}.attn.bias',
        f"h.{idx}.attn.c_attn.weight",
        f"h.{idx}.attn.c_attn.bias",
        f"h.{idx}.attn.c_proj.weight",
        f"h.{idx}.attn.c_proj.bias",
        f"h.{idx}.ln_2.weight",
        f"h.{idx}.ln_2.bias",
        f"h.{idx}.mlp.c_fc.weight",
        f"h.{idx}.mlp.c_fc.bias",
        f"h.{idx}.mlp.c_proj.weight",
        f"h.{idx}.mlp.c_proj.bias",
    ]
    if idx == config.n_layers - 1:
        names += ["ln_f.weight", "ln_f.bias"]
    _names = ["transformer." + name for name in names]
    return names + _names


def gpt2_output_head_param_name(config: ReaLModelConfig) -> List[str]:
    return ["wte.weight", "transformer.wte.weight", "lm_head.weight"]


def convert_config_gpt2(
    hf_config: transformers.GPT2Config,
) -> ReaLModelConfig:
    return ReaLModelConfig(
        n_layers=hf_config.n_layer,
        n_kv_heads=hf_config.n_head,
        hidden_dim=hf_config.n_embd,
        head_dim=hf_config.n_embd // hf_config.n_head,
        n_q_heads=hf_config.n_head,
        intermediate_dim=(
            hf_config.n_inner if hf_config.n_inner is not None else 4 * hf_config.n_embd
        ),
        vocab_size=hf_config.vocab_size,
        n_positions=hf_config.n_positions,
        embd_pdrop=hf_config.embd_pdrop,
        attn_pdrop=hf_config.attn_pdrop,
        resid_pdrop=hf_config.resid_pdrop,
        layer_norm_epsilon=hf_config.layer_norm_epsilon,
        activation_function=hf_config.activation_function,
        use_attention_bias=True,
        use_attn_proj_bias=True,
        scale_attn_by_inverse_layer_idx=hf_config.scale_attn_by_inverse_layer_idx,
        tied_embedding=True,
        scale_attn_weights=hf_config.scale_attn_weights,
    )


def to_gpt2_config(config: ReaLModelConfig) -> transformers.GPT2Config:
    return transformers.GPT2Config(
        vocab_size=config.vocab_size,
        n_positions=config.n_positions,
        n_embd=config.hidden_dim,
        n_layer=config.n_layers,
        n_head=config.n_q_heads,
        n_inner=config.intermediate_dim,
        activation_function=config.activation_function,
        embd_pdrop=config.embd_pdrop,
        attn_pdrop=config.attn_pdrop,
        resid_pdrop=config.resid_pdrop,
        layer_norm_epsilon=config.layer_norm_epsilon,
        scale_attn_by_inverse_layer_idx=config.scale_attn_by_inverse_layer_idx,
        scale_attn_weights=config.scale_attn_weights,
    )


def make_real_config_gpt2() -> ReaLModelConfig:
    hf_config = transformers.GPT2Config(
        vocab_size=160,
        n_positions=200,
        n_embd=128,
        n_layer=8,
        n_head=16,
        n_inner=512,
        activation_function="gelu_new",
    )
    return convert_config_gpt2(hf_config)


register_hf_family(
    name="gpt2",
    hf_cls_name="GPT2LMHeadModel",
    config_from_hf_converter=convert_config_gpt2,
    config_to_hf_converter=to_gpt2_config,
    sd_from_hf_converter=sd_from_gpt2,
    sd_to_hf_converter=sd_to_gpt2,
    embedding_param_names=gpt2_embedding_layer_names,
    tblock_param_names=gpt2_transformer_block_param_name,
    head_param_names=gpt2_output_head_param_name,
    real_config_maker=make_real_config_gpt2,
)



import dataclasses
import itertools
import json
import os
import pathlib
import shutil
import uuid
from typing import *

import pytest
import torch
import torch.distributed as dist

from realhf.api.core.config import ModelFamily, ModelName, ModelShardID
from realhf.api.core.model_api import HF_MODEL_FAMILY_REGISTRY, ReaLModelConfig
from realhf.base import constants, logging, topology
from realhf.base.testing import (
    LocalMultiProcessTest,
    clear_name_resolve,
    init_global_constants,
    make_random_packed_batches,
)

if TYPE_CHECKING:
    from realhf.impl.model.backend.megatron import ReaLMegatronEngine
    from realhf.impl.model.nn.real_llm_api import ReaLModel


def compute_critic_loss(
    logits: torch.Tensor,
    packed_input_ids: torch.Tensor,
    cu_seqlens: torch.Tensor,
    prompt_mask: torch.Tensor,
    **kwargs,
) -> torch.Tensor:
    from realhf.impl.model.utils.functional import (
        build_shift_one_indices,
        gather_packed_shifted_log_probs,
    )

    shift_one_indices = build_shift_one_indices(logits, cu_seqlens)
    prompt_mask = prompt_mask[shift_one_indices]
    scores = logits.squeeze().float()[shift_one_indices]
    scores = torch.where(prompt_mask, 0, scores)

    loss = ((scores - torch.zeros_like(scores)) ** 2).sum() / (
        prompt_mask.numel() - prompt_mask.count_nonzero()
    )
    return loss, {"loss": loss.clone().detach()}


def create_model(
    tmp_dir: pathlib.Path,
    model_family_name: str,
    model_name,
    is_critic: int,
    instantiate=True,
) -> "ReaLModel":
    # NOTE: import here to avoid initializing CUDA context in the main process
    from realhf.impl.model.nn.real_llm_api import ReaLModel

    with constants.model_scope(model_name):
        mconfig: ReaLModelConfig = getattr(
            ReaLModel, f"make_{model_family_name}_config"
        )()
        mconfig.is_critic = is_critic

        # initialize model
        model = ReaLModel(mconfig, dtype=torch.float16, device="cuda")
        model.eval()
        if instantiate:
            model.instantiate()
    if instantiate:
        _check_tied_embedding_weights(model_name, model)
    with constants.model_scope(model_name):
        if instantiate:
            init_save_path = tmp_dir / "init"
            # sync initialized parameters
            getattr(model, f"to_{model_family_name}")(None, init_save_path)
            dist.barrier(group=constants.parallelism_group())
            model = getattr(model, f"from_{model_family_name}")(
                init_save_path, init_critic_from_actor=False
            )
    if instantiate:
        _check_tied_embedding_weights(model_name, model)
    return model


def get_topo(model_name):
    with constants.model_scope(model_name):
        return constants.grid().topology()


def build_engine(module, model_name, trainable) -> "ReaLMegatronEngine":
    from realhf.api.core import model_api
    from realhf.impl.model.backend.inference import PipelineInferenceBackend
    from realhf.impl.model.backend.megatron import MegatronTrainBackend
    from realhf.impl.model.nn.real_llm_api import add_helper_functions

    with constants.model_scope(model_name):
        if constants.pipe_parallel_world_size() == 1:
            add_helper_functions(module)
        if trainable:
            backend = MegatronTrainBackend(initial_loss_scale=8.0)
        else:
            backend = PipelineInferenceBackend()
        _model = backend.initialize(
            model_api.Model(
                None,
                module,
                None,
                module.device,
                module.dtype,
            ),
            model_api.FinetuneSpec(
                total_train_epochs=1,
                total_train_steps=20,
                steps_per_epoch=20,
            ),
        )
    return _model.module


def setup_constants_and_param_realloc(
    from_model_name,
    to_model_name,
    from_pp_dp_mp,
    to_pp_dp_mp,
    from_sequence_parallel,
    to_sequence_parallel,
):
    from realhf.impl.model.comm.param_realloc import set_trainable, setup_param_realloc

    from_num_pp, from_num_dp, from_num_mp = from_pp_dp_mp
    to_num_pp, to_num_dp, to_num_mp = to_pp_dp_mp

    from_world_size = from_num_dp * from_num_mp * from_num_pp
    to_world_size = to_num_dp * to_num_mp * to_num_pp

    from_topo = topology.PipeModelDataParallelTopology(
        num_dp=from_num_dp,
        num_mp=from_num_mp,
        num_pp=from_num_pp,
        sequence_parallel=from_sequence_parallel,
        gradient_checkpointing=False,
        max_prompt_len=None,
    )
    to_topo = topology.PipeModelDataParallelTopology(
        num_dp=to_num_dp,
        num_mp=to_num_mp,
        num_pp=to_num_pp,
        sequence_parallel=to_sequence_parallel,
        gradient_checkpointing=False,
        max_prompt_len=None,
    )

    model_topos = {from_model_name: from_topo, to_model_name: to_topo}

    msid2mwid = {}
    for i in range(dist.get_world_size()):
        # We assume the `from_model` occupies the first serveral GPUs,
        # while the `to_model` occupies GPUs from the last one.
        # For example, when the world size of `from_model` is 6 and
        # the world size of `to_model` is 4, the GPU layout is:
        # GPU 0-3: from_model (shard 0-3)
        # GPU 4-5: from_model (shard 4-5) + to_model (shard 0-1)
        # GPU 6-7: to_model (shard 2-3)
        _model_names = []
        if i < from_world_size:
            _model_names.append(from_model_name)
        if i >= dist.get_world_size() - to_world_size:
            _model_names.append(to_model_name)
        for _model_name in _model_names:
            if _model_name == from_model_name:
                coord = model_topos[_model_name].get_coord(i)
            else:
                coord = model_topos[_model_name].get_coord(
                    i + to_world_size - dist.get_world_size()
                )
            k = ModelShardID(
                _model_name,
                dp_rank=coord.data,
                mp_rank=coord.model,
                pp_rank=coord.pipe,
                topo=model_topos[_model_name],
            )
            msid2mwid[k] = i

    init_global_constants(
        num_dp=from_num_dp,
        num_mp=from_num_mp,
        num_pp=from_num_pp,
        topo=from_topo,
        model_name=from_model_name,
        sequence_parallel=from_sequence_parallel,
        msid2mwid=msid2mwid,
    )

    init_global_constants(
        num_dp=to_num_dp,
        num_mp=to_num_mp,
        num_pp=to_num_pp,
        model_name=to_model_name,
        sequence_parallel=to_sequence_parallel,
        msid2mwid=msid2mwid,
    )

    pg_info = setup_param_realloc(
        model_topos=model_topos,
        msid2mwid=msid2mwid,
        param_realloc_pairs=[
            (from_model_name, to_model_name),
            (to_model_name, from_model_name),
        ],
    )
    set_trainable(from_model_name, True)
    set_trainable(to_model_name, False)
    return pg_info


def _check_tied_embedding_weights(model_name, model: "ReaLModel"):
    with constants.model_scope(model_name):
        if not (constants.is_first_pipe_stage() or constants.is_last_pipe_stage()):
            return

        if constants.is_first_pipe_stage():
            w1 = w = model.layers[0].wte.weight
        if constants.is_last_pipe_stage():
            w2 = w = model.layers[-1].weight

        if constants.pipe_parallel_world_size() == 1:
            if model.config.tied_embedding and not model.config.is_critic:
                assert w1.data_ptr() == w2.data_ptr()
            else:
                assert w1.data_ptr() != w2.data_ptr()
        else:
            w_ = w.clone().detach()
            dist.all_reduce(
                w_,
                op=dist.ReduceOp.SUM,
                group=constants.grid().embedding_proc_group,
            )
            w_ /= dist.get_world_size(constants.grid().embedding_proc_group)
            if model.config.tied_embedding and not model.config.is_critic:
                assert torch.allclose(w_, w), (w_ - w).abs().max()
            else:
                assert not torch.allclose(w_, w), (w_ - w).abs().max()


@dataclasses.dataclass
class ParamRedistributer:
    from_model_name: ModelName
    to_model_name: ModelName
    from_model: Any
    to_model: Any
    pg_info: Any

    def _redist(self, m1, m2, n1, n2):
        if m1 is None and m2 is None:
            return
        with constants.model_scope(n1):
            t1 = constants.grid().topology()
        with constants.model_scope(n2):
            t2 = constants.grid().topology()
        m = m1 if m1 is not None else m2
        a, b, c = m.build_reparallelized_layers_async(
            from_model_name=n1,
            to_model_name=n2,
            from_topo=t1,
            to_topo=t2,
            to_model_config=m.config,
            pg_info=self.pg_info,
        )
        if m2 is not None:
            m2.patch_reparallelization((a, b))

        if m1 is not None:
            assert m1.layers is None
            assert m1.contiguous_param is None
        if m2 is not None:
            assert m2.layers is not None
            assert m2.contiguous_param is not None

    def forward(self):
        self._redist(
            self.from_model,
            self.to_model,
            self.from_model_name,
            self.to_model_name,
        )

    def backward(self):
        self._redist(
            self.to_model,
            self.from_model,
            self.to_model_name,
            self.from_model_name,
        )


def _load_all_pytorch_bin(path: pathlib.Path):
    if os.path.exists(path / "pytorch_model.bin.index.json"):
        with open(path / "pytorch_model.bin.index.json", "r") as f:
            hf_sd_mapping = json.load(f)["weight_map"]
        sd = {}
        for fn in hf_sd_mapping.values():
            sd.update(torch.load(path / fn, map_location="cpu"))
    else:
        sd = torch.load(path / "pytorch_model.bin", map_location="cpu")
    return sd


def _test_para_realloc(
    tmp_path: pathlib.Path,
    model_family_name: str,
    is_critic: bool,
    from_pp_dp_mp: Tuple,
    to_pp_dp_mp: Tuple,
    from_sequence_parallel: bool,
    to_sequence_parallel: bool,
    n_iterations: int,
    skip_saveload: bool,
):
    # os.environ["REAL_SAVE_MAX_SHARD_SIZE_BYTE"] = str(int(1e6))
    from realhf.impl.model.backend.megatron import ReaLMegatronEngine
    from realhf.impl.model.interface.sft_interface import compute_packed_sft_loss

    from_model_name = ModelName("param_realloc_test", 0)
    to_model_name = ModelName("param_realloc_test", 1)

    pg_info = setup_constants_and_param_realloc(
        from_model_name,
        to_model_name,
        from_pp_dp_mp,
        to_pp_dp_mp,
        from_sequence_parallel,
        to_sequence_parallel,
    )

    # Create model 1
    if dist.get_rank() < from_pp_dp_mp[0] * from_pp_dp_mp[1] * from_pp_dp_mp[2]:
        from_model = create_model(
            tmp_dir=tmp_path,
            model_family_name=model_family_name,
            model_name=from_model_name,
            is_critic=is_critic,
            instantiate=True,
        )
    else:
        from_model = None
    # Creat model 2
    if (
        dist.get_rank()
        >= dist.get_world_size() - to_pp_dp_mp[0] * to_pp_dp_mp[1] * to_pp_dp_mp[2]
    ):
        to_model = create_model(
            tmp_dir=tmp_path,
            model_family_name=model_family_name,
            model_name=to_model_name,
            is_critic=is_critic,
            instantiate=False,
        )
    else:
        to_model = None

    # Create redistributer.
    redist = ParamRedistributer(
        from_model_name,
        to_model_name,
        from_model,
        to_model,
        pg_info,
    )

    if from_model is not None:
        train_engine = build_engine(from_model, from_model_name, trainable=True)
        _check_tied_embedding_weights(from_model_name, from_model)
    if to_model is not None:
        inf_engine = build_engine(to_model, to_model_name, trainable=False)

    for i in range(n_iterations):
        # Create the same random data across all ranks.
        if from_model is not None:
            vocab_size = from_model.config.vocab_size
        elif to_model is not None:
            vocab_size = to_model.config.vocab_size
        else:
            # Give a random vocab size for sampling across the whole world.
            vocab_size = 100

        _v = torch.tensor([vocab_size], dtype=torch.int32, device="cuda")
        dist.all_reduce(_v, op=dist.ReduceOp.MAX)
        vocab_size = _v.item()

        # Synchronize the data across all ranks.
        x = make_random_packed_batches(
            n_batches=1,
            batch_size=32,
            seq_len=32,
            vocab_size=vocab_size,
            dp_rank=0,
            dp_size=1,
        )[0]
        packed_input_ids = x["packed_prompts"].cuda()
        dist.all_reduce(packed_input_ids, op=dist.ReduceOp.MAX)
        input_lens = torch.tensor(x.metadata["seqlens"], dtype=torch.int, device="cuda")
        cu_seqlens = torch.nn.functional.pad(input_lens.cumsum(0), (1, 0)).int()
        prompt_mask = torch.zeros_like(packed_input_ids, dtype=torch.bool)
        assert torch.all(prompt_mask == 0)
        input_lens = cu_seqlens[1:] - cu_seqlens[:-1]
        seqlens_cpu = input_lens.cpu().numpy().tolist()

        # Synchronize the initial parameters at the start of this iteration.
        if not skip_saveload:
            if from_model is not None:
                with constants.model_scope(from_model_name):
                    getattr(from_model, f"to_{model_family_name}")(
                        None, tmp_path / f"save_from_{i}"
                    )
            dist.barrier()
            sd1 = _load_all_pytorch_bin(tmp_path / f"save_from_{i}")

        # Run redistribution.
        redist.forward()
        dist.barrier()

        # Synchronize the redistributed parameters. They should be identical to the initial parameters.
        # Also, they should be different from the parameters of the previous iteration
        # because we have updated the parameters.
        if not skip_saveload:
            if to_model is not None:
                with constants.model_scope(to_model_name):
                    getattr(to_model, f"to_{model_family_name}")(
                        None, tmp_path / f"save_to_{i}"
                    )
            dist.barrier()
            sd2 = _load_all_pytorch_bin(tmp_path / f"save_to_{i}")
            for k, v in sd1.items():
                assert torch.allclose(v, sd2[k]), (k, v, sd2[k])

        # Run a forward with the redistributed model.
        if to_model is not None:
            _check_tied_embedding_weights(to_model_name, to_model)
            with constants.model_scope(to_model_name):
                inf_engine.eval()
                logits1 = inf_engine.forward(
                    seqlens_cpu=seqlens_cpu,
                    packed_input_ids=packed_input_ids,
                    cu_seqlens=cu_seqlens,
                )

        # Run redistribution backwards.
        redist.backward()
        dist.barrier()

        # Re-run redistribution to examine whether inference results are identical.
        redist.forward()
        dist.barrier()
        if to_model is not None:
            _check_tied_embedding_weights(to_model_name, to_model)
            with constants.model_scope(to_model_name):
                inf_engine.eval()
                logits2 = inf_engine.forward(
                    seqlens_cpu=seqlens_cpu,
                    packed_input_ids=packed_input_ids,
                    cu_seqlens=cu_seqlens,
                )
            if logits1 is not None:
                assert torch.allclose(logits1, logits2)
        redist.backward()
        dist.barrier()

        # Synchronize the redistributed parameters. They should be identical to the initial parameters.
        if not skip_saveload:
            if from_model is not None:
                with constants.model_scope(from_model_name):
                    getattr(from_model, f"to_{model_family_name}")(
                        None, tmp_path / f"save_back_{i}"
                    )
            dist.barrier()
            sd3 = _load_all_pytorch_bin(tmp_path / f"save_back_{i}")
            for k, v in sd1.items():
                assert torch.allclose(v, sd3[k]), (k, v, sd3[k])

        # Train the model.
        if from_model is not None:
            for _ in range(2):
                _check_tied_embedding_weights(from_model_name, from_model)
                loss_fn_kwargs = dict(
                    prompt_mask=prompt_mask,
                    input_lens=input_lens,
                )
                train_engine.eval()

                p = from_model.contiguous_param.clone().detach()

                with constants.model_scope(from_model_name):
                    train_engine: ReaLMegatronEngine
                    stats = train_engine.train_batch(
                        seqlens_cpu=seqlens_cpu,
                        packed_input_ids=packed_input_ids,
                        cu_seqlens=cu_seqlens,
                        loss_fn=(
                            compute_packed_sft_loss
                            if not is_critic
                            else compute_critic_loss
                        ),
                        num_micro_batches=constants.pipe_parallel_world_size() * 2,
                        version_steps=i,
                        **loss_fn_kwargs,
                    )

                p_ = from_model.contiguous_param.clone().detach()
                # After training, the parameters should have changed.
                assert not torch.allclose(p, p_), (p - p_).abs().max()

        # Re-run redistribution to ensure that inference results changed.
        redist.forward()
        dist.barrier()
        if to_model is not None:
            _check_tied_embedding_weights(to_model_name, to_model)
            with constants.model_scope(to_model_name):
                inf_engine.eval()
                logits3 = inf_engine.forward(
                    seqlens_cpu=seqlens_cpu,
                    packed_input_ids=packed_input_ids,
                    cu_seqlens=cu_seqlens,
                )
            if logits1 is not None:
                assert not torch.allclose(logits1, logits3)
        redist.backward()
        dist.barrier()
    print("success")


parallelism = [(4, 1, 1), (2, 2, 2), (1, 8, 1)]


@pytest.mark.skipif(
    not torch.cuda.is_available() or torch.cuda.device_count() < 8,
    reason="This test requires at least 8 GPUs to run.",
)
@pytest.mark.parametrize("model_family_name", ["gpt2"])
@pytest.mark.parametrize("is_critic", [False])
@pytest.mark.parametrize("from_pp_dp_mp", [(4, 1, 1)])
@pytest.mark.parametrize("to_pp_dp_mp", [(2, 2, 2)])
@pytest.mark.parametrize("from_sequence_parallel", [False])
@pytest.mark.parametrize("to_sequence_parallel", [False])
@pytest.mark.parametrize("skip_saveload", [True])
@pytest.mark.slow
@pytest.mark.distributed
def test_param_realloc(
    tmp_path: pathlib.Path,
    model_family_name: str,
    is_critic: bool,
    from_pp_dp_mp: Tuple,
    to_pp_dp_mp: Tuple,
    from_sequence_parallel: bool,
    to_sequence_parallel: bool,
    skip_saveload: bool,
):
    if from_sequence_parallel and from_pp_dp_mp[-1] == 1:
        return
    if to_sequence_parallel and to_pp_dp_mp[-1] == 1:
        return
    expr_name = uuid.uuid4()
    trial_name = uuid.uuid4()
    test_impl = LocalMultiProcessTest(
        world_size=8,
        func=_test_para_realloc,
        expr_name=expr_name,
        trial_name=trial_name,
        tmp_path=tmp_path,
        model_family_name=model_family_name,
        is_critic=is_critic,
        from_pp_dp_mp=from_pp_dp_mp,
        to_pp_dp_mp=to_pp_dp_mp,
        from_sequence_parallel=from_sequence_parallel,
        to_sequence_parallel=to_sequence_parallel,
        n_iterations=4,
        skip_saveload=skip_saveload,
    )
    test_impl.launch()



import json
import os
import pathlib
import shutil
import uuid
from typing import *

import pytest
import torch
import torch.distributed as dist
import transformers

from realhf.api.core.config import ModelFamily
from realhf.api.core.model_api import HF_MODEL_FAMILY_REGISTRY, ReaLModelConfig
from realhf.base import constants, logging
from realhf.base.testing import (
    LocalMultiProcessTest,
    clear_name_resolve,
    init_global_constants,
)

logger = logging.getLogger("tests.test_saveload")


def _load_all_pytorch_bin(path: pathlib.Path):
    if os.path.exists(path / "pytorch_model.bin.index.json"):
        with open(path / "pytorch_model.bin.index.json", "r") as f:
            hf_sd_mapping = json.load(f)["weight_map"]
        sd = {}
        for fn in hf_sd_mapping.values():
            sd.update(torch.load(path / fn, map_location="cpu"))
    else:
        sd = torch.load(path / "pytorch_model.bin", map_location="cpu")
    return sd


def _save_then_load(
    tmp_path: pathlib.Path,
    model_family_name: str,
    is_critic: bool,
    init_critic_from_actor: bool,
    pp_dp_mp: Tuple,
):
    # NOTE: import here to avoid initializing CUDA context in the main process
    from realhf.impl.model.nn.real_llm_api import ReaLModel

    # os.environ["REAL_SAVE_MAX_SHARD_SIZE_BYTE"] = str(int(1e6))

    model_name = f"saveload_test_{model_family_name}"
    num_pp, num_dp, num_mp = pp_dp_mp
    init_global_constants(
        num_dp=num_dp,
        num_mp=num_mp,
        num_pp=num_pp,
        model_name=model_name,
    )
    assert dist.get_world_size() == 8, dist.get_world_size()
    assert tmp_path.exists()
    init_save_path = tmp_path / "init"
    real_save_path = tmp_path / "real"
    real_save_path2 = tmp_path / "real2"

    with constants.model_scope(model_name):
        tokenizer = None
        mconfig: ReaLModelConfig = getattr(
            ReaLModel, f"make_{model_family_name}_config"
        )()
        mconfig.is_critic = is_critic

        # load from hf model or create a new critic model
        model = ReaLModel(mconfig, dtype=torch.float32, device="cuda")
        model.instantiate()
        # sync initialized parameters
        getattr(model, f"to_{model_family_name}")(tokenizer, init_save_path)
        dist.barrier()
        model = getattr(model, f"from_{model_family_name}")(
            init_save_path, init_critic_from_actor=False
        )
        sd1 = model.state_dict()

        # save ReaLModel (e.g., after SFT)
        getattr(model, f"to_{model_family_name}")(tokenizer, real_save_path)
        dist.barrier()
        file_size = 0
        for fn in os.listdir(real_save_path):
            if fn.endswith(".bin"):
                file_size += os.path.getsize(os.path.join(real_save_path, fn))

        # load ReaLModel (e.g., before PPO, RW)
        model = ReaLModel(mconfig, dtype=torch.float32, device="cuda")
        model._instantiation_hooks.append(
            lambda: getattr(model, f"from_{model_family_name}")(
                real_save_path, init_critic_from_actor
            )
        )
        model.instantiate()
        dist.barrier()
        sd2 = model.state_dict()
        for k, v in sd2.items():
            if init_critic_from_actor and k == f"{mconfig.n_layers + 1}.weight":
                continue
            assert torch.allclose(v, sd1[k]), (
                k,
                v.flatten()[:10],
                sd1[k].flatten()[:10],
            )

        # Load saved ReaLModel using HF APIs.
        if not is_critic:
            hf_model = transformers.AutoModelForCausalLM.from_pretrained(real_save_path)
            dist.barrier()
            _hf_sd = hf_model.state_dict()
            sd3 = _load_all_pytorch_bin(real_save_path)
            if model_family_name != "gpt2":
                for k, v in sd3.items():
                    if k.endswith(".rotary_emb.inv_freq"):
                        continue
                    assert torch.allclose(v.cpu(), _hf_sd[k]), k
            else:
                for k, v in sd3.items():
                    if k.endswith(".attn.bias"):
                        continue
                    assert torch.allclose(v.cpu(), _hf_sd[f"transformer.{k}"]), k

        # save again, check size
        getattr(model, f"to_{model_family_name}")(tokenizer, real_save_path2)
        dist.barrier()
        file_size2 = 0
        for fn in os.listdir(real_save_path2):
            if fn.endswith(".bin"):
                file_size2 += os.path.getsize(os.path.join(real_save_path2, fn))
        assert file_size2 == file_size, (file_size, file_size2)
        dist.barrier()


@pytest.mark.skipif(
    not torch.cuda.is_available() or torch.cuda.device_count() < 8,
    reason="This test requires at least 8 GPUs to run.",
)
@pytest.mark.slow
@pytest.mark.distributed
@pytest.mark.parametrize("model_family_name", ["gemma", "gpt2", "llama", "qwen2", "mistral"])
@pytest.mark.parametrize("is_critic", [True, False])
@pytest.mark.parametrize("init_critic_from_actor", [True, False])
@pytest.mark.parametrize("pp_dp_mp", [(4, 2, 1), (2, 2, 2), (1, 2, 4), (1, 8, 1)])
def test_save_then_load(
    tmp_path: pathlib.Path,
    model_family_name: str,
    is_critic: bool,
    init_critic_from_actor: bool,
    pp_dp_mp: Tuple,
):
    if model_family_name == "gpt2" and pp_dp_mp[-1] > 1:
        return
    if not is_critic and init_critic_from_actor:
        return
    expr_name = uuid.uuid4()
    trial_name = uuid.uuid4()
    test_impl = LocalMultiProcessTest(
        world_size=8,
        func=_save_then_load,
        expr_name=expr_name,
        trial_name=trial_name,
        model_family_name=model_family_name,
        is_critic=is_critic,
        init_critic_from_actor=init_critic_from_actor,
        pp_dp_mp=pp_dp_mp,
        tmp_path=tmp_path,
    )
    test_impl.launch()
